# Copyright (c) 2014 Clinton Knight.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# Copyright (c) 2016 Michael Price.  All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Mock unit tests for the NetApp driver utility module
"""

import copy
import ddt
import platform

import mock
from oslo_concurrency import processutils as putils

from cinder import context
from cinder import exception
from cinder import test
from cinder.tests.unit.volume.drivers.netapp.dataontap.client import (
    fakes as zapi_fakes)
import cinder.tests.unit.volume.drivers.netapp.fakes as fake
from cinder import version
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp import utils as na_utils
from cinder.volume import qos_specs
from cinder.volume import volume_types


@ddt.ddt
class NetAppDriverUtilsTestCase(test.TestCase):

    @mock.patch.object(na_utils, 'LOG', mock.Mock())
    def test_validate_instantiation_proxy(self):
        kwargs = {'netapp_mode': 'proxy'}
        na_utils.validate_instantiation(**kwargs)
        self.assertEqual(0, na_utils.LOG.warning.call_count)

    @mock.patch.object(na_utils, 'LOG', mock.Mock())
    def test_validate_instantiation_no_proxy(self):
        kwargs = {'netapp_mode': 'asdf'}
        na_utils.validate_instantiation(**kwargs)
        self.assertEqual(1, na_utils.LOG.warning.call_count)

    def test_check_flags(self):

        class TestClass(object):
            pass

        required_flags = ['flag1', 'flag2']
        configuration = TestClass()
        setattr(configuration, 'flag1', 'value1')
        setattr(configuration, 'flag3', 'value3')
        self.assertRaises(exception.InvalidInput, na_utils.check_flags,
                          required_flags, configuration)

        setattr(configuration, 'flag2', 'value2')
        self.assertIsNone(na_utils.check_flags(required_flags, configuration))

    def test_to_bool(self):
        self.assertTrue(na_utils.to_bool(True))
        self.assertTrue(na_utils.to_bool('true'))
        self.assertTrue(na_utils.to_bool('yes'))
        self.assertTrue(na_utils.to_bool('y'))
        self.assertTrue(na_utils.to_bool(1))
        self.assertTrue(na_utils.to_bool('1'))
        self.assertFalse(na_utils.to_bool(False))
        self.assertFalse(na_utils.to_bool('false'))
        self.assertFalse(na_utils.to_bool('asdf'))
        self.assertFalse(na_utils.to_bool('no'))
        self.assertFalse(na_utils.to_bool('n'))
        self.assertFalse(na_utils.to_bool(0))
        self.assertFalse(na_utils.to_bool('0'))
        self.assertFalse(na_utils.to_bool(2))
        self.assertFalse(na_utils.to_bool('2'))

    def test_set_safe_attr(self):

        fake_object = mock.Mock()
        fake_object.fake_attr = None

        # test initial checks
        self.assertFalse(na_utils.set_safe_attr(None, fake_object, None))
        self.assertFalse(na_utils.set_safe_attr(fake_object, None, None))
        self.assertFalse(na_utils.set_safe_attr(fake_object, 'fake_attr',
                                                None))

        # test value isn't changed if it shouldn't be and retval is False
        fake_object.fake_attr = 'fake_value'
        self.assertFalse(na_utils.set_safe_attr(fake_object, 'fake_attr',
                                                'fake_value'))
        self.assertEqual('fake_value', fake_object.fake_attr)

        # test value is changed if it should be and retval is True
        self.assertTrue(na_utils.set_safe_attr(fake_object, 'fake_attr',
                                               'new_fake_value'))
        self.assertEqual('new_fake_value', fake_object.fake_attr)

    def test_round_down(self):
        self.assertAlmostEqual(na_utils.round_down(5.567), 5.56)
        self.assertAlmostEqual(na_utils.round_down(5.567, '0.00'), 5.56)
        self.assertAlmostEqual(na_utils.round_down(5.567, '0.0'), 5.5)
        self.assertAlmostEqual(na_utils.round_down(5.567, '0'), 5)
        self.assertAlmostEqual(na_utils.round_down(0, '0.00'), 0)
        self.assertAlmostEqual(na_utils.round_down(-5.567), -5.56)
        self.assertAlmostEqual(na_utils.round_down(-5.567, '0.00'), -5.56)
        self.assertAlmostEqual(na_utils.round_down(-5.567, '0.0'), -5.5)
        self.assertAlmostEqual(na_utils.round_down(-5.567, '0'), -5)

    def test_iscsi_connection_properties(self):
        actual_properties = na_utils.get_iscsi_connection_properties(
            fake.ISCSI_FAKE_LUN_ID, fake.ISCSI_FAKE_VOLUME,
            [fake.ISCSI_FAKE_IQN, fake.ISCSI_FAKE_IQN2],
            [fake.ISCSI_FAKE_ADDRESS_IPV4, fake.ISCSI_FAKE_ADDRESS2_IPV4],
            [fake.ISCSI_FAKE_PORT, fake.ISCSI_FAKE_PORT])

        actual_properties_mapped = actual_properties['data']

        self.assertDictEqual(actual_properties_mapped,
                             fake.ISCSI_MP_TARGET_INFO_DICT)

    def test_iscsi_connection_properties_single_iqn(self):
        actual_properties = na_utils.get_iscsi_connection_properties(
            fake.ISCSI_FAKE_LUN_ID, fake.ISCSI_FAKE_VOLUME,
            fake.ISCSI_FAKE_IQN,
            [fake.ISCSI_FAKE_ADDRESS_IPV4, fake.ISCSI_FAKE_ADDRESS2_IPV4],
            [fake.ISCSI_FAKE_PORT, fake.ISCSI_FAKE_PORT])

        actual_properties_mapped = actual_properties['data']
        expected = fake.ISCSI_MP_TARGET_INFO_DICT.copy()
        expected['target_iqns'][1] = expected['target_iqns'][0]

        self.assertDictEqual(actual_properties_mapped,
                             fake.ISCSI_MP_TARGET_INFO_DICT)

    def test_iscsi_connection_lun_id_type_str(self):
        FAKE_LUN_ID = '1'

        actual_properties = na_utils.get_iscsi_connection_properties(
            FAKE_LUN_ID, fake.ISCSI_FAKE_VOLUME, fake.ISCSI_FAKE_IQN,
            [fake.ISCSI_FAKE_ADDRESS_IPV4], [fake.ISCSI_FAKE_PORT])

        actual_properties_mapped = actual_properties['data']

        self.assertIs(int, type(actual_properties_mapped['target_lun']))
        self.assertDictEqual(actual_properties_mapped,
                             fake.FC_ISCSI_TARGET_INFO_DICT)

    def test_iscsi_connection_lun_id_type_dict(self):
        FAKE_LUN_ID = {'id': 'fake_id'}

        self.assertRaises(TypeError, na_utils.get_iscsi_connection_properties,
                          FAKE_LUN_ID, fake.ISCSI_FAKE_VOLUME,
                          fake.ISCSI_FAKE_IQN, [fake.ISCSI_FAKE_ADDRESS_IPV4],
                          [fake.ISCSI_FAKE_PORT])

    def test_iscsi_connection_properties_ipv6(self):
        actual_properties = na_utils.get_iscsi_connection_properties(
            '1', fake.ISCSI_FAKE_VOLUME_NO_AUTH, fake.ISCSI_FAKE_IQN,
            [fake.ISCSI_FAKE_ADDRESS_IPV6], [fake.ISCSI_FAKE_PORT])

        self.assertDictEqual(actual_properties['data'],
                             fake.FC_ISCSI_TARGET_INFO_DICT_IPV6)

    def test_get_volume_extra_specs(self):
        fake_extra_specs = {'fake_key': 'fake_value'}
        fake_volume_type = {'extra_specs': fake_extra_specs}
        fake_volume = {'volume_type_id': 'fake_volume_type_id'}
        self.mock_object(context, 'get_admin_context')
        self.mock_object(volume_types, 'get_volume_type',
                         return_value=fake_volume_type)
        self.mock_object(na_utils, 'log_extra_spec_warnings')

        result = na_utils.get_volume_extra_specs(fake_volume)

        self.assertEqual(fake_extra_specs, result)

    def test_trace_filter_func_api(self):
        na_utils.setup_api_trace_pattern("^(?!(perf)).*$")
        na_element = zapi_fakes.FAKE_NA_ELEMENT
        all_args = {'na_element': na_element}
        self.assertTrue(na_utils.trace_filter_func_api(all_args))

    def test_trace_filter_func_api_invalid(self):
        all_args = {'fake': 'not_na_element'}
        self.assertTrue(na_utils.trace_filter_func_api(all_args))

    def test_trace_filter_func_api_filtered(self):
        na_utils.setup_api_trace_pattern("^(?!(perf)).*$")
        na_element = netapp_api.NaElement("perf-object-counter-list-info")
        all_args = {'na_element': na_element}
        self.assertFalse(na_utils.trace_filter_func_api(all_args))

    def test_get_volume_extra_specs_no_type_id(self):
        fake_volume = {}
        self.mock_object(context, 'get_admin_context')
        self.mock_object(volume_types, 'get_volume_type')
        self.mock_object(na_utils, 'log_extra_spec_warnings')

        result = na_utils.get_volume_extra_specs(fake_volume)

        self.assertEqual({}, result)

    def test_get_volume_extra_specs_no_volume_type(self):
        fake_volume = {'volume_type_id': 'fake_volume_type_id'}
        self.mock_object(context, 'get_admin_context')
        self.mock_object(volume_types, 'get_volume_type', return_value=None)
        self.mock_object(na_utils, 'log_extra_spec_warnings')

        result = na_utils.get_volume_extra_specs(fake_volume)

        self.assertEqual({}, result)

    def test_log_extra_spec_warnings_obsolete_specs(self):

        mock_log = self.mock_object(na_utils.LOG, 'warning')

        na_utils.log_extra_spec_warnings({'netapp:raid_type': 'raid4'})

        self.assertEqual(1, mock_log.call_count)

    def test_log_extra_spec_warnings_deprecated_specs(self):

        mock_log = self.mock_object(na_utils.LOG, 'warning')

        na_utils.log_extra_spec_warnings({'netapp_thick_provisioned': 'true'})

        self.assertEqual(1, mock_log.call_count)

    def test_validate_qos_spec_none(self):
        qos_spec = None

        # Just return without raising an exception.
        na_utils.validate_qos_spec(qos_spec)

    def test_validate_qos_spec_keys_weirdly_cased(self):
        qos_spec = {'mAxIopS': 33000}

        # Just return without raising an exception.
        na_utils.validate_qos_spec(qos_spec)

    def test_validate_qos_spec_bad_key(self):
        qos_spec = {'maxFlops': 33000}

        self.assertRaises(exception.Invalid,
                          na_utils.validate_qos_spec,
                          qos_spec)

    def test_validate_qos_spec_bad_key_combination(self):
        qos_spec = {'maxIOPS': 33000, 'maxBPS': 10000000}

        self.assertRaises(exception.Invalid,
                          na_utils.validate_qos_spec,
                          qos_spec)

    def test_map_qos_spec_none(self):
        qos_spec = None

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertIsNone(result)

    def test_map_qos_spec_maxiops(self):
        qos_spec = {'maxIOPs': 33000}
        mock_get_name = self.mock_object(na_utils, 'get_qos_policy_group_name')
        mock_get_name.return_value = 'fake_qos_policy'
        expected = {
            'policy_name': 'fake_qos_policy',
            'max_throughput': '33000iops',
        }

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertEqual(expected, result)

    def test_map_qos_spec_maxiopspergib(self):
        qos_spec = {'maxIOPSperGiB': 1000}
        mock_get_name = self.mock_object(na_utils, 'get_qos_policy_group_name')
        mock_get_name.return_value = 'fake_qos_policy'
        expected = {
            'policy_name': 'fake_qos_policy',
            'max_throughput': '42000iops',
        }

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertEqual(expected, result)

    def test_map_qos_spec_maxbps(self):
        qos_spec = {'maxBPS': 1000000}
        mock_get_name = self.mock_object(na_utils, 'get_qos_policy_group_name')
        mock_get_name.return_value = 'fake_qos_policy'
        expected = {
            'policy_name': 'fake_qos_policy',
            'max_throughput': '1000000B/s',
        }

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertEqual(expected, result)

    def test_map_qos_spec_maxbpspergib(self):
        qos_spec = {'maxBPSperGiB': 100000}
        mock_get_name = self.mock_object(na_utils, 'get_qos_policy_group_name')
        mock_get_name.return_value = 'fake_qos_policy'
        expected = {
            'policy_name': 'fake_qos_policy',
            'max_throughput': '4200000B/s',
        }

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertEqual(expected, result)

    def test_map_qos_spec_no_key_present(self):
        qos_spec = {}
        mock_get_name = self.mock_object(na_utils, 'get_qos_policy_group_name')
        mock_get_name.return_value = 'fake_qos_policy'
        expected = {
            'policy_name': 'fake_qos_policy',
            'max_throughput': None,
        }

        result = na_utils.map_qos_spec(qos_spec, fake.VOLUME)

        self.assertEqual(expected, result)

    def test_map_dict_to_lower(self):
        original = {'UPperKey': 'Value'}
        expected = {'upperkey': 'Value'}

        result = na_utils.map_dict_to_lower(original)

        self.assertEqual(expected, result)

    def test_get_qos_policy_group_name(self):
        expected = 'openstack-%s' % fake.VOLUME_ID

        result = na_utils.get_qos_policy_group_name(fake.VOLUME)

        self.assertEqual(expected, result)

    def test_get_qos_policy_group_name_no_id(self):
        volume = copy.deepcopy(fake.VOLUME)
        del(volume['id'])

        result = na_utils.get_qos_policy_group_name(volume)

        self.assertIsNone(result)

    def test_get_qos_policy_group_name_from_info(self):
        expected = 'openstack-%s' % fake.VOLUME_ID
        result = na_utils.get_qos_policy_group_name_from_info(
            fake.QOS_POLICY_GROUP_INFO)

        self.assertEqual(expected, result)

    def test_get_qos_policy_group_name_from_info_no_info(self):

        result = na_utils.get_qos_policy_group_name_from_info(None)

        self.assertIsNone(result)

    def test_get_qos_policy_group_name_from_legacy_info(self):
        expected = fake.QOS_POLICY_GROUP_NAME

        result = na_utils.get_qos_policy_group_name_from_info(
            fake.LEGACY_QOS_POLICY_GROUP_INFO)

        self.assertEqual(expected, result)

    def test_get_qos_policy_group_name_from_spec_info(self):
        expected = 'openstack-%s' % fake.VOLUME_ID

        result = na_utils.get_qos_policy_group_name_from_info(
            fake.QOS_POLICY_GROUP_INFO)

        self.assertEqual(expected, result)

    def test_get_qos_policy_group_name_from_none_qos_info(self):
        expected = None

        result = na_utils.get_qos_policy_group_name_from_info(
            fake.QOS_POLICY_GROUP_INFO_NONE)

        self.assertEqual(expected, result)

    def test_get_valid_qos_policy_group_info_exception_path(self):
        mock_get_volume_type = self.mock_object(na_utils,
                                                'get_volume_type_from_volume')
        mock_get_volume_type.side_effect = exception.VolumeTypeNotFound
        expected = fake.QOS_POLICY_GROUP_INFO_NONE

        result = na_utils.get_valid_qos_policy_group_info(fake.VOLUME)

        self.assertEqual(expected, result)

    def test_get_valid_qos_policy_group_info_volume_type_none(self):
        mock_get_volume_type = self.mock_object(na_utils,
                                                'get_volume_type_from_volume')
        mock_get_volume_type.return_value = None
        expected = fake.QOS_POLICY_GROUP_INFO_NONE

        result = na_utils.get_valid_qos_policy_group_info(fake.VOLUME)

        self.assertEqual(expected, result)

    def test_get_valid_qos_policy_group_info_no_info(self):
        mock_get_volume_type = self.mock_object(na_utils,
                                                'get_volume_type_from_volume')
        mock_get_volume_type.return_value = fake.VOLUME_TYPE
        mock_get_legacy_qos_policy = self.mock_object(na_utils,
                                                      'get_legacy_qos_policy')
        mock_get_legacy_qos_policy.return_value = None
        mock_get_valid_qos_spec_from_volume_type = self.mock_object(
            na_utils, 'get_valid_backend_qos_spec_from_volume_type')
        mock_get_valid_qos_spec_from_volume_type.return_value = None
        self.mock_object(na_utils, 'check_for_invalid_qos_spec_combination')
        expected = fake.QOS_POLICY_GROUP_INFO_NONE

        result = na_utils.get_valid_qos_policy_group_info(fake.VOLUME)

        self.assertEqual(expected, result)

    def test_get_valid_legacy_qos_policy_group_info(self):
        mock_get_volume_type = self.mock_object(na_utils,
                                                'get_volume_type_from_volume')
        mock_get_volume_type.return_value = fake.VOLUME_TYPE
        mock_get_legacy_qos_policy = self.mock_object(na_utils,
                                                      'get_legacy_qos_policy')

        mock_get_legacy_qos_policy.return_value = fake.LEGACY_QOS
        mock_get_valid_qos_spec_from_volume_type = self.mock_object(
            na_utils, 'get_valid_backend_qos_spec_from_volume_type')
        mock_get_valid_qos_spec_from_volume_type.return_value = None
        self.mock_object(na_utils, 'check_for_invalid_qos_spec_combination')

        result = na_utils.get_valid_qos_policy_group_info(fake.VOLUME)

        self.assertEqual(fake.LEGACY_QOS_POLICY_GROUP_INFO, result)

    def test_get_valid_spec_qos_policy_group_info(self):
        mock_get_volume_type = self.mock_object(na_utils,
                                                'get_volume_type_from_volume')
        mock_get_volume_type.return_value = fake.VOLUME_TYPE
        mock_get_legacy_qos_policy = self.mock_object(na_utils,
                                                      'get_legacy_qos_policy')
        mock_get_legacy_qos_policy.return_value = None
        mock_get_valid_qos_spec_from_volume_type = self.mock_object(
            na_utils, 'get_valid_backend_qos_spec_from_volume_type')
        mock_get_valid_qos_spec_from_volume_type.return_value =\
            fake.QOS_POLICY_GROUP_SPEC
        self.mock_object(na_utils, 'check_for_invalid_qos_spec_combination')

        result = na_utils.get_valid_qos_policy_group_info(fake.VOLUME)

        self.assertEqual(fake.QOS_POLICY_GROUP_INFO, result)

    def test_get_valid_backend_qos_spec_from_volume_type_no_spec(self):
        mock_get_spec = self.mock_object(
            na_utils, 'get_backend_qos_spec_from_volume_type')
        mock_get_spec.return_value = None
        mock_validate = self.mock_object(na_utils, 'validate_qos_spec')

        result = na_utils.get_valid_backend_qos_spec_from_volume_type(
            fake.VOLUME, fake.VOLUME_TYPE)

        self.assertIsNone(result)
        self.assertEqual(0, mock_validate.call_count)

    def test_get_valid_backend_qos_spec_from_volume_type(self):
        mock_get_spec = self.mock_object(
            na_utils, 'get_backend_qos_spec_from_volume_type')
        mock_get_spec.return_value = fake.QOS_SPEC
        mock_validate = self.mock_object(na_utils, 'validate_qos_spec')

        result = na_utils.get_valid_backend_qos_spec_from_volume_type(
            fake.VOLUME, fake.VOLUME_TYPE)

        self.assertEqual(fake.QOS_POLICY_GROUP_SPEC, result)
        self.assertEqual(1, mock_validate.call_count)

    def test_get_backend_qos_spec_from_volume_type_no_qos_specs_id(self):
        volume_type = copy.deepcopy(fake.VOLUME_TYPE)
        del(volume_type['qos_specs_id'])
        mock_get_context = self.mock_object(context, 'get_admin_context')

        result = na_utils.get_backend_qos_spec_from_volume_type(volume_type)

        self.assertIsNone(result)
        self.assertEqual(0, mock_get_context.call_count)

    def test_get_backend_qos_spec_from_volume_type_no_qos_spec(self):
        volume_type = fake.VOLUME_TYPE
        self.mock_object(context, 'get_admin_context')
        mock_get_specs = self.mock_object(qos_specs, 'get_qos_specs')
        mock_get_specs.return_value = None

        result = na_utils.get_backend_qos_spec_from_volume_type(volume_type)

        self.assertIsNone(result)

    def test_get_backend_qos_spec_from_volume_type_with_frontend_spec(self):
        volume_type = fake.VOLUME_TYPE
        self.mock_object(context, 'get_admin_context')
        mock_get_specs = self.mock_object(qos_specs, 'get_qos_specs')
        mock_get_specs.return_value = fake.OUTER_FRONTEND_QOS_SPEC

        result = na_utils.get_backend_qos_spec_from_volume_type(volume_type)

        self.assertIsNone(result)

    def test_get_backend_qos_spec_from_volume_type_with_backend_spec(self):
        volume_type = fake.VOLUME_TYPE
        self.mock_object(context, 'get_admin_context')
        mock_get_specs = self.mock_object(qos_specs, 'get_qos_specs')
        mock_get_specs.return_value = fake.OUTER_BACKEND_QOS_SPEC

        result = na_utils.get_backend_qos_spec_from_volume_type(volume_type)

        self.assertEqual(fake.QOS_SPEC, result)

    def test_get_backend_qos_spec_from_volume_type_with_both_spec(self):
        volume_type = fake.VOLUME_TYPE
        self.mock_object(context, 'get_admin_context')
        mock_get_specs = self.mock_object(qos_specs, 'get_qos_specs')
        mock_get_specs.return_value = fake.OUTER_BOTH_QOS_SPEC

        result = na_utils.get_backend_qos_spec_from_volume_type(volume_type)

        self.assertEqual(fake.QOS_SPEC, result)

    def test_check_for_invalid_qos_spec_combination(self):

        self.assertRaises(exception.Invalid,
                          na_utils.check_for_invalid_qos_spec_combination,
                          fake.INVALID_QOS_POLICY_GROUP_INFO,
                          fake.VOLUME_TYPE)

    def test_get_legacy_qos_policy(self):
        extra_specs = fake.LEGACY_EXTRA_SPECS
        expected = {'policy_name': fake.QOS_POLICY_GROUP_NAME}

        result = na_utils.get_legacy_qos_policy(extra_specs)

        self.assertEqual(expected, result)

    def test_get_legacy_qos_policy_no_policy_name(self):
        extra_specs = fake.EXTRA_SPECS

        result = na_utils.get_legacy_qos_policy(extra_specs)

        self.assertIsNone(result)

    @ddt.data(("192.168.99.24:/fake/export/path", "192.168.99.24",
               "/fake/export/path"),
              ("127.0.0.1:/", "127.0.0.1", "/"),
              ("[f180::30d9]:/path_to-export/3.1/this folder", "f180::30d9",
               "/path_to-export/3.1/this folder"),
              ("[::]:/", "::", "/"),
              ("[2001:db8::1]:/fake_export", "2001:db8::1", "/fake_export"))
    @ddt.unpack
    def test_get_export_host_junction_path(self, share, host, junction_path):
        result_host, result_path = na_utils.get_export_host_junction_path(
            share)

        self.assertEqual(host, result_host)
        self.assertEqual(junction_path, result_path)

    @ddt.data("192.14.21.0/wrong_export", "192.14.21.0:8080:/wrong_export"
              "2001:db8::1:/wrong_export",
              "[2001:db8::1:/wrong_export", "2001:db8::1]:/wrong_export")
    def test_get_export_host_junction_path_with_invalid_exports(self, share):
        self.assertRaises(na_utils.NetAppDriverException,
                          na_utils.get_export_host_junction_path,
                          share)


class OpenStackInfoTestCase(test.TestCase):

    UNKNOWN_VERSION = 'unknown version'
    UNKNOWN_RELEASE = 'unknown release'
    UNKNOWN_VENDOR = 'unknown vendor'
    UNKNOWN_PLATFORM = 'unknown platform'
    VERSION_STRING_RET_VAL = 'fake_version_1'
    RELEASE_STRING_RET_VAL = 'fake_release_1'
    PLATFORM_RET_VAL = 'fake_platform_1'
    VERSION_INFO_VERSION = 'fake_version_2'
    VERSION_INFO_RELEASE = 'fake_release_2'
    RPM_INFO_VERSION = 'fake_version_3'
    RPM_INFO_RELEASE = 'fake_release_3'
    RPM_INFO_VENDOR = 'fake vendor 3'
    PUTILS_RPM_RET_VAL = ('fake_version_3  fake_release_3 fake vendor 3', '')
    NO_PKG_FOUND = ('', 'whatever')
    PUTILS_DPKG_RET_VAL = ('epoch:upstream_version-debian_revision', '')
    DEB_RLS = 'upstream_version-debian_revision'
    DEB_VENDOR = 'debian_revision'

    def test_openstack_info_init(self):
        info = na_utils.OpenStackInfo()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(version.version_info, 'version_string',
                       mock.Mock(return_value=VERSION_STRING_RET_VAL))
    def test_update_version_from_version_string(self):
        info = na_utils.OpenStackInfo()
        info._update_version_from_version_string()

        self.assertEqual(self.VERSION_STRING_RET_VAL, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(version.version_info, 'version_string',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_update_version_from_version_string(self):
        info = na_utils.OpenStackInfo()
        info._update_version_from_version_string()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(version.version_info, 'release_string',
                       mock.Mock(return_value=RELEASE_STRING_RET_VAL))
    def test_update_release_from_release_string(self):
        info = na_utils.OpenStackInfo()
        info._update_release_from_release_string()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.RELEASE_STRING_RET_VAL, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(version.version_info, 'release_string',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_update_release_from_release_string(self):
        info = na_utils.OpenStackInfo()
        info._update_release_from_release_string()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(platform, 'platform',
                       mock.Mock(return_value=PLATFORM_RET_VAL))
    def test_update_platform(self):
        info = na_utils.OpenStackInfo()
        info._update_platform()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.PLATFORM_RET_VAL, info._platform)

    @mock.patch.object(platform, 'platform',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_update_platform(self):
        info = na_utils.OpenStackInfo()
        info._update_platform()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_version',
                       mock.Mock(return_value=VERSION_INFO_VERSION))
    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_release',
                       mock.Mock(return_value=VERSION_INFO_RELEASE))
    def test_update_info_from_version_info(self):
        info = na_utils.OpenStackInfo()
        info._update_info_from_version_info()

        self.assertEqual(self.VERSION_INFO_VERSION, info._version)
        self.assertEqual(self.VERSION_INFO_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_version',
                       mock.Mock(return_value=''))
    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_release',
                       mock.Mock(return_value=None))
    def test_no_info_from_version_info(self):
        info = na_utils.OpenStackInfo()
        info._update_info_from_version_info()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_version',
                       mock.Mock(return_value=VERSION_INFO_VERSION))
    @mock.patch.object(na_utils.OpenStackInfo, '_get_version_info_release',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_info_from_version_info(self):
        info = na_utils.OpenStackInfo()
        info._update_info_from_version_info()

        self.assertEqual(self.VERSION_INFO_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(return_value=PUTILS_RPM_RET_VAL))
    def test_update_info_from_rpm(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_rpm()

        self.assertEqual(self.RPM_INFO_VERSION, info._version)
        self.assertEqual(self.RPM_INFO_RELEASE, info._release)
        self.assertEqual(self.RPM_INFO_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertTrue(found_package)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(return_value=NO_PKG_FOUND))
    def test_update_info_from_rpm_no_pkg_found(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_rpm()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertFalse(found_package)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_update_info_from_rpm(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_rpm()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertFalse(found_package)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(return_value=PUTILS_DPKG_RET_VAL))
    def test_update_info_from_dpkg(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_dpkg()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.DEB_RLS, info._release)
        self.assertEqual(self.DEB_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertTrue(found_package)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(return_value=NO_PKG_FOUND))
    def test_update_info_from_dpkg_no_pkg_found(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_dpkg()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertFalse(found_package)

    @mock.patch.object(putils, 'execute',
                       mock.Mock(side_effect=Exception))
    def test_xcption_in_update_info_from_dpkg(self):
        info = na_utils.OpenStackInfo()
        found_package = info._update_info_from_dpkg()

        self.assertEqual(self.UNKNOWN_VERSION, info._version)
        self.assertEqual(self.UNKNOWN_RELEASE, info._release)
        self.assertEqual(self.UNKNOWN_VENDOR, info._vendor)
        self.assertEqual(self.UNKNOWN_PLATFORM, info._platform)
        self.assertFalse(found_package)

    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_version_from_version_string', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_release_from_release_string', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_platform', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_version_info', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_rpm', mock.Mock(return_value=True))
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_dpkg')
    def test_update_openstack_info_rpm_pkg_found(self, mock_updt_from_dpkg):
        info = na_utils.OpenStackInfo()
        info._update_openstack_info()

        self.assertFalse(mock_updt_from_dpkg.called)

    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_version_from_version_string', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_release_from_release_string', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_platform', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_version_info', mock.Mock())
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_rpm', mock.Mock(return_value=False))
    @mock.patch.object(na_utils.OpenStackInfo,
                       '_update_info_from_dpkg')
    def test_update_openstack_info_rpm_pkg_not_found(self,
                                                     mock_updt_from_dpkg):
        info = na_utils.OpenStackInfo()
        info._update_openstack_info()

        self.assertTrue(mock_updt_from_dpkg.called)


@ddt.ddt
class FeaturesTestCase(test.TestCase):

    def setUp(self):
        super(FeaturesTestCase, self).setUp()
        self.features = na_utils.Features()

    def test_init(self):
        self.assertSetEqual(set(), self.features.defined_features)

    def test_add_feature_default(self):
        self.features.add_feature('FEATURE_1')

        self.assertTrue(self.features.FEATURE_1.supported)
        self.assertIn('FEATURE_1', self.features.defined_features)

    @ddt.data(True, False)
    def test_add_feature(self, value):
        self.features.add_feature('FEATURE_2', value)

        self.assertEqual(value, bool(self.features.FEATURE_2))
        self.assertEqual(value, self.features.FEATURE_2.supported)
        self.assertIsNone(self.features.FEATURE_2.minimum_version)
        self.assertIn('FEATURE_2', self.features.defined_features)

    @ddt.data((True, '1'), (False, 2), (False, None), (True, None))
    @ddt.unpack
    def test_add_feature_min_version(self, enabled, min_version):
        self.features.add_feature('FEATURE_2', enabled,
                                  min_version=min_version)

        self.assertEqual(enabled, bool(self.features.FEATURE_2))
        self.assertEqual(enabled, self.features.FEATURE_2.supported)
        self.assertEqual(min_version, self.features.FEATURE_2.minimum_version)
        self.assertIn('FEATURE_2', self.features.defined_features)

    @ddt.data('True', 'False', 0, 1, 1.0, None, [], {}, (True,))
    def test_add_feature_type_error(self, value):
        self.assertRaises(TypeError,
                          self.features.add_feature,
                          'FEATURE_3',
                          value)
        self.assertNotIn('FEATURE_3', self.features.defined_features)

    def test_get_attr_missing(self):
        self.assertRaises(AttributeError, getattr, self.features, 'FEATURE_4')


@ddt.ddt
class BitSetTestCase(test.TestCase):

    def test_default(self):
        self.assertEqual(na_utils.BitSet(0), na_utils.BitSet())

    def test_set(self):
        bitset = na_utils.BitSet(0)
        bitset.set(16)

        self.assertEqual(na_utils.BitSet(1 << 16), bitset)

    def test_unset(self):
        bitset = na_utils.BitSet(1 << 16)
        bitset.unset(16)

        self.assertEqual(na_utils.BitSet(0), bitset)

    def test_is_set(self):
        bitset = na_utils.BitSet(1 << 16)

        self.assertTrue(bool(bitset.is_set(16)))

    def test_not_equal(self):
        set1 = na_utils.BitSet(1 << 15)
        set2 = na_utils.BitSet(1 << 16)

        self.assertNotEqual(set1, set2)

    def test_repr(self):
        raw_val = 1 << 16
        actual = repr(na_utils.BitSet(raw_val))
        expected = str(raw_val)

        self.assertEqual(actual, expected)

    def test_str(self):
        raw_val = 1 << 16
        actual = str(na_utils.BitSet(raw_val))
        expected = bin(raw_val)

        self.assertEqual(actual, expected)

    def test_int(self):
        val = 1 << 16
        actual = int(int(na_utils.BitSet(val)))

        self.assertEqual(val, actual)

    def test_and(self):
        actual = na_utils.BitSet(1 << 16 | 1 << 15)
        actual &= 1 << 16

        self.assertEqual(na_utils.BitSet(1 << 16), actual)

    def test_or(self):
        actual = na_utils.BitSet()
        actual |= 1 << 16

        self.assertEqual(na_utils.BitSet(1 << 16), actual)

    def test_invert(self):
        actual = na_utils.BitSet(1 << 16)
        actual = ~actual

        self.assertEqual(~(1 << 16), actual)

    def test_xor(self):
        actual = na_utils.BitSet(1 << 16)
        actual ^= 1 << 16

        self.assertEqual(na_utils.BitSet(), actual)

    def test_lshift(self):
        actual = na_utils.BitSet(1)
        actual <<= 16

        self.assertEqual(na_utils.BitSet(1 << 16), actual)

    def test_rshift(self):
        actual = na_utils.BitSet(1 << 16)
        actual >>= 16

        self.assertEqual(na_utils.BitSet(1), actual)



# Copyright (c) 2018 Hedvig, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Rest Client for Hedvig Openstack implementation.
"""
import json
import random

from oslo_log import log as logging
from oslo_utils import units
from six.moves import http_client
from six.moves import urllib

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.hedvig import config

LOG = logging.getLogger(__name__)


class RestClient(object):
    def __init__(self, nodes, username, password, cluster):
        """Hedvig Rest Client

        :param node: hostname of one of the nodes in the cluster
        :param username:  username of the cluster
        :param password:  password of the cluster
        :param cluster:  clustername of the cluster
        """
        LOG.debug('init called with %s , %s', nodes, cluster)
        self.username = username
        self.password = password
        self.cluster = cluster
        self.nodes = nodes
        self.nodeMap = {}

    def connect(self):
        self.store_node_map(self.nodes)
        if len(self.nodeMap) is 0:
            msg = _('Unable to connect to the nodes')
            raise exception.VolumeDriverException(msg)

    def get_session_id(self, node):
        """Retrieves the session Id

        :param node: hostname of the node
        :return:  session ID which is valid for 15 minutes
        """
        LOG.debug("get_session_id called with node %s", node)
        data = {
            'request': {
                'type': 'Login',
                'category': 'UserManagement',
                'params': {
                    'userName': self.username,
                    'password': self.password,
                    'cluster': self.cluster
                }
            }
        }
        obj = self.query(data=data, node=node)
        if obj['status'] != 'ok':
            msg = _('GetSessionId failure')
            raise exception.VolumeDriverException(msg)
        return (obj['result']['sessionId']).encode('utf-8')

    def get_all_cluster_nodes(self, node):
        """Retrieves all the nodes present in the cluster

        :param node: hostname of the node
        :return:  nodes present in the cluster
        """
        LOG.debug("get_all_cluster_nodes called with node %s", node)
        data = {
            'request': {
                'type': 'ListClusterNodes',
                'category': 'VirtualDiskManagement',
                'sessionId': self.get_session_id(node),
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        return obj['result']

    def store_node_map(self, nodes):
        """Stores all the node information along with their sessionID in dict

        :param nodes: hostname of the nodes in the cluster
        """
        LOG.debug("store_node_map called with node %s", nodes)
        exitFlag = False
        node_list = []
        for n in nodes.split(','):
            node_list.append(n.strip())
        for node in node_list:
            try:
                LOG.debug("Attempting store_node_map with node %s", node)
                nodeList = self.get_all_cluster_nodes(node)
                exitFlag = True
                for node_ in nodeList:
                    self.nodeMap[node_] = self.get_session_id(node_)

            except urllib.error.HTTPError as e:
                if e.code == http_client.NOT_FOUND:
                    LOG.debug("Client not found")
                else:
                    LOG.debug("Client not available")

            except Exception:
                LOG.exception('Retrying store_node_map with next node')

            if exitFlag:
                return

    def refresh_session_ids(self):
        """In case of session failure , it  refreshes all the

        session ID stored in nodeMap
        """
        LOG.debug("refresh_session_ids called")
        if len(self.nodeMap.keys()) == 0:
            msg = _('NodeMap is empty')
            raise exception.VolumeDriverException(msg)
        for node, val in self.nodeMap.items():
            self.nodeMap[node] = self.get_session_id(node)

    def query(self, data, node):
        """Makes a rest query with given params

        :param data: json given as param to Rest call
        :param node:  hostname of the node
        :return: REST response
        """
        data = urllib.parse.urlencode(data)
        req = urllib.request.Request("http://%s/rest/" % node, data)
        response = urllib.request.urlopen(req)
        json_str = response.read()
        obj = json.loads(json_str)
        LOG.debug("Rest call output %s ", json_str)
        return obj

    def make_rest_call(self, data, node):
        """Makes a rest Call and retries it 5 times in case of rest failure

        :param data: json given as param to Rest call
        :param node: hostname of the node
        :return:
        """
        retryCount = 0
        while retryCount < config.Config.retryCount:
            retryCount = retryCount + 1
            try:
                LOG.debug("Rest call started with node %s "
                          "and data: %s", node, data)
                obj = self.query(data, node)
                if obj['status'] == 'ok' or obj['status'] == 'warning':
                    return obj
                # We need to refresh sessionIDs if earlier ones are expired
                elif 'session-failure' in obj['status']:
                    self.refresh_session_ids()
                    session_id = self.retrieve_session_id(node)
                    data['request']['sessionId'] = session_id

            except Exception as e:
                LOG.debug("Exception details: data - %s, node - %s "
                          "exception - %s", data, node, e.args)
                node = self.get_pages_host()
        else:
            msg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(msg)

    def create_vdisk(self, vDiskInfo):
        """Rest call to create a vdisk

        :param vDiskInfo: json passsed to the rest call
        """
        LOG.debug("create_vdisk called")
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)

        sizeInB = vDiskInfo['size'] / units.Gi

        sizeInJson = {'unit': "GB",
                      'value': float(sizeInB)}
        vDiskInfo['size'] = sizeInJson
        data = {
            'request': {
                'type': 'AddVirtualDisk',
                'category': 'VirtualDiskManagement',
                'params': vDiskInfo,
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        if obj['result'][0]['status'] != 'ok':
            errmsg = _('create_vdisk REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def resize_vdisk(self, vDiskName, value):
        """Rest Call to resize Vdisk

        :param vDiskName:  name of the vdisk
        :param unit: unit is GB for openstack
        :param value:  size of the resized vdisk in GB
        """
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        LOG.debug("resize_vdisk called")
        data = {
            'request': {
                'type': 'ResizeDisks',
                'category': 'VirtualDiskManagement',
                'params': {
                    'virtualDisks': [vDiskName.encode('utf-8')],
                    'size': {
                        'unit': "GB",
                        'value': value
                    },
                },
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)

        if obj['result'][0]['status'] != 'ok':
            errmsg = _('resize_vdisk REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def delete_vdisk(self, vDiskName):
        """Rest call to delete Vdisk

        :param vDiskName: name of the vdisk
        :return: Status of the rest call
        """
        LOG.debug("delete_vdisk called %s", vDiskName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'DeleteVDisk',
                'category': 'VirtualDiskManagement',
                'params': {
                    'virtualDisks': [vDiskName.encode('utf-8')],
                },
                'sessionId': sessionId,
            }
        }

        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok':
            if "couldn't be found" not in obj['message']:
                errmsg = _('REST call status - %s') % obj['status']
                raise exception.VolumeDriverException(errmsg)

    def get_lun(self, target, vDiskName):
        """Retrieve lun number

        :param target: hostname of the target
        :param vDiskName: name of the Vdisk
        :return: lun number
        """
        try:
            LOG.debug("get_lun called for vdisk %s", vDiskName)
            node = self.get_pages_host()
            sessionId = self.retrieve_session_id(node)
            data = {
                'request': {
                    'type': 'GetLun',
                    'category': 'VirtualDiskManagement',
                    'params': {
                        'virtualDisk': vDiskName.encode('utf-8'),
                        'target': target.encode('utf-8'),
                    },
                    'sessionId': sessionId,
                }
            }

            obj = self.make_rest_call(data=data, node=node)

            if obj['status'] != 'ok':
                return -1
            return obj['result']['lun']
        except Exception:
            return -1

    def get_iqn(self, host):
        """Retrieve IQN of the host.

        :param host: hostname
        :return: iqn of the host
        """
        LOG.debug("get_iqn called for host %s", host)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'GetIqn',
                'category': 'VirtualDiskManagement',
                'params': {
                    'host': host.encode('utf-8'),
                },
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok':
            if "IQN not found" in obj['message']:
                return "ALL"
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)
        return obj['result']['iqn']

    def add_lun(self, tgtHost, vDiskName, readonly):
        """Rest Call to Add Lun

        :param tgtHost: hostname of target
        :param vDiskName: name of vdisk
        :param readonly: boolean readonly value
        """
        LOG.debug(
            "add_lun called with target %s, vdisk %s", tgtHost, vDiskName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'AddLun',
                'category': 'VirtualDiskManagement',
                'params': {
                    'virtualDisks': [vDiskName.encode('utf-8')],
                    'targets': [tgtHost.encode('utf-8')],
                    'readonly': readonly,
                },
                'sessionId': sessionId,
            }
        }

        obj = self.make_rest_call(data=data, node=node)

        restCallStatus = obj['result'][0]['status']
        tgts = obj['result'][0]['targets']
        addLunStatus = tgts[0]['status']
        if restCallStatus != 'ok' or addLunStatus != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def unmap_lun(self, target, vDiskName):
        """Rest call to unmap Lun

        :param target: hostname of the target
        :param vDiskName: name of the vdisk
        :return: true if successful
        """
        LOG.debug("unmap_lun called with target %s, vdisk %s", target,
                  vDiskName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'UnmapLun',
                'category': 'VirtualDiskManagement',
                'params': {
                    'virtualDisk': vDiskName.encode('utf-8'),
                    'target': target.encode('utf-8'),
                },
                'sessionId': sessionId,
            }
        }

        obj = self.make_rest_call(data=data, node=node)

        if obj['status'] != 'ok':
            msg = "is not mapped to the specified controller"
            if(msg not in obj['message']):
                errmsg = _('REST call status - %s') % obj['status']
                raise exception.VolumeDriverException(errmsg)

        return True

    def add_access(self, host, vDiskName, type, address):
        """Rest Call to Add access

        :param host: hostname
        :param vDiskName: name of vdisk
        :param type: type is iqn for openstack
        :param address: iqn address
        """
        LOG.debug(
            "add_access called with param host %s, vdisk %s",
            host, vDiskName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'PersistACLAccess',
                'category': 'VirtualDiskManagement',
                'params': {
                    'virtualDisks': [vDiskName.encode('utf-8')],
                    'host': host.encode('utf-8'),
                    'type': type.encode('utf-8'),
                    'address': address.encode('utf-8')
                },
                'sessionId': sessionId,
            }
        }

        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok' or obj['result'][0]['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def create_snapshot(self, vDiskName, snapshotId):
        """Rest Call to create snapshot

        :param vDiskName: name of the vdisk
        :param snapshotId: snapshotId of the snapshot
        :return: status of the rest call
        """
        LOG.debug("create_snapshot called with vdisk %s", vDiskName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'MakeSnapshot',
                'category': 'SnapshotManagement',
                'params': {
                    'virtualDisks': [vDiskName.encode('utf-8')],
                },
                'sessionId': sessionId,
            }
        }
        if snapshotId:
            param = data['request']['params']
            param['openstackSID'] = snapshotId.encode('utf-8')
        obj = self.make_rest_call(data=data, node=node)

        if obj['status'] != 'ok' or obj['result'][0]['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

        return obj['result'][0]['snapshotName'].encode('utf-8')

    def clone_vdisk(self, srcVolName, dstVolName, size):
        """Rest Call to clone vdisk

        """
        LOG.debug("clonevdisk called vdisk %s, %s", srcVolName, dstVolName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'CloneVdisk',
                'category': 'SnapshotManagement',
                'params': {
                    'srcVolName': srcVolName.encode('utf-8'),
                    'cloneVolName': dstVolName.encode('utf-8'),
                    'size': size
                },
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def get_val_in_gb(self, value, unit):
        unitRef = {
            'B': 1,
            'KB': units.Ki,
            'MB': units.Mi,
            'GB': units.Gi,
            'TB': units.Ti,
            'PB': units.Pi
        }
        return value * unitRef[unit] / units.Gi

    def update_volume_stats(self):
        """Fetch cluster level details"""
        LOG.debug("Update volume stats called")
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'ClusterInformation',
                'category': 'ClusterWatch',
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

        total = obj['result']['capacity']['total']['value']
        used = obj['result']['capacity']['used']['value']
        capacity = obj['result']['capacity']
        total_unit = capacity['total']['units'].encode('utf-8')
        used_unit = capacity['used']['units'].encode('utf-8')
        total_capacity = self.get_val_in_gb(total, total_unit)
        used_capacity = self.get_val_in_gb(used, used_unit)
        free_capacity = total_capacity - used_capacity
        LOG.debug("total_capacity %s free_capactity %s", total_capacity,
                  free_capacity)
        return (total_capacity, free_capacity)

    def clone_hedvig_snapshot(self, dstVolName, snapshotID, srcVolName, size):
        """Rest Call to clone hedvig snapshot

        """
        LOG.debug("clone_hedvig_snapshot %s, %s", dstVolName, srcVolName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'CloneVdisk',
                'category': 'SnapshotManagement',
                'params': {
                    'cloneVolName': dstVolName.encode('utf-8'),
                    'openstackSID': snapshotID.encode('utf-8'),
                    'srcVolName': srcVolName.encode('utf-8'),
                    'size': size
                },
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)
        if obj['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def delete_snapshot(self, snapshotName, vDiskName, snapshotId):
        """Rest call to delete snapshot

        :param snapshotName:  name of the snapshot to be deleted
        """
        LOG.debug(
            "delete_snapshot called with snapshot %s", snapshotName)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        data = {
            'request': {
                'type': 'DeleteSnapshot',
                'category': 'SnapshotManagement',
                'params': {
                    'snapshotName': snapshotName.encode('utf-8'),
                    'openstackSID': snapshotId.encode('utf-8'),
                    'openstackVolName': vDiskName.encode('utf-8')
                },
                'sessionId': sessionId,
            }
        }
        obj = self.make_rest_call(data=data, node=node)

        if obj['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

    def list_targets(self, computeHost):
        """Rest Call to ListTargets for a given hostname

        :param computeHost:  hostname of the computeHost
        :return: list of targets
        """
        LOG.debug("list_targets called with computehost %s", computeHost)
        node = self.get_pages_host()
        sessionId = self.retrieve_session_id(node)
        targets = []
        data = {
            'request': {
                'type': 'ListTargets',
                'category': 'VirtualDiskManagement',
                'sessionId': sessionId,

            }
        }
        if computeHost:
            data['request']['params'] = {}
            data['request']['params']['computeHost'] = computeHost

        obj = self.make_rest_call(data=data, node=node)

        if obj['status'] != 'ok':
            errmsg = _('REST call status - %s') % obj['status']
            raise exception.VolumeDriverException(errmsg)

        for ch in obj['result']:
            if ch['protocol'] == 'block':
                targets.append(ch['target'])

        return targets

    def get_pages_host(self):
        """Returns a random host from nodemap

        :return:  hostname
        """
        LOG.debug("get_pages_host called")
        if not self.nodeMap:
            msg = _('NodeMap is empty')
            raise exception.VolumeDriverException(msg)
        return random.choice(self.nodeMap.keys())

    def retrieve_session_id(self, node):
        """returns sessionID of the given node

        :param node: hostname of the node
        :return:  session ID of  the given host
        """
        LOG.debug("retrieve_session_id called with node %s", node)
        if len(self.nodeMap.keys()) == 0:
            msg = _('NodeMap is empty')
            raise exception.VolumeDriverException(msg)
        return self.nodeMap[str(node)]



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg

from cinder.volume import configuration

CONF = cfg.CONF

CONF.import_opt('policy_file', 'cinder.policy', group='oslo_policy')
CONF.import_opt('volume_driver', 'cinder.volume.manager',
                group=configuration.SHARED_CONF_GROUP)
CONF.import_opt('backup_driver', 'cinder.backup.manager')
CONF.import_opt('backend', 'cinder.keymgr', group='key_manager')
CONF.import_opt('scheduler_driver', 'cinder.scheduler.manager')

def_vol_type = '__DEFAULT__'


def set_defaults(conf):
    conf.set_default('default_volume_type', def_vol_type)
    conf.set_default('volume_driver',
                     'cinder.tests.fake_driver.FakeLoggingVolumeDriver',
                     group=configuration.SHARED_CONF_GROUP)
    conf.set_default('target_helper', 'fake')
    conf.set_default('transport_url', 'fake:/')
    conf.set_default('connection', 'sqlite://', group='database')
    conf.set_default('sqlite_synchronous', False, group='database')
    conf.set_default('policy_file', 'cinder.tests.unit/policy.json',
                     group='oslo_policy')
    conf.set_default('backup_driver',
                     'cinder.tests.unit.backup.fake_service.FakeBackupService')
    conf.set_default('backend',
                     'castellan.tests.unit.key_manager.mock_key_manager.'
                     'MockKeyManager',
                     group='key_manager')
    conf.set_default('scheduler_driver',
                     'cinder.scheduler.filter_scheduler.FilterScheduler')
    conf.set_default('state_path', os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', '..', '..')))
    conf.set_default('policy_dirs', [], group='oslo_policy')
    # This is where we don't authenticate
    conf.set_default('auth_strategy', 'noauth')
    conf.set_default('auth_url', 'fake', 'keystone_authtoken')



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for group API.
"""

import ddt
import mock

from cinder import context
from cinder import exception
import cinder.group
from cinder import objects
from cinder.objects import fields
from cinder.policies import group_snapshots as g_snap_policies
from cinder import quota
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import utils


GROUP_QUOTAS = quota.GROUP_QUOTAS


@ddt.ddt
class GroupAPITestCase(test.TestCase):
    """Test Case for group API."""

    def setUp(self):
        super(GroupAPITestCase, self).setUp()
        self.group_api = cinder.group.API()
        self.ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                           auth_token=True,
                                           is_admin=True)
        self.user_ctxt = context.RequestContext(
            fake.USER_ID, fake.PROJECT_ID, auth_token=True)

    @mock.patch('cinder.objects.Group.get_by_id')
    def test_get(self, mock_group_get):
        fake_group = {'name': 'fake_group'}
        mock_group_get.return_value = fake_group
        grp = self.group_api.get(self.ctxt, fake.GROUP_ID)
        self.assertEqual(fake_group, grp)

    @ddt.data(True, False)
    @mock.patch('cinder.objects.GroupList.get_all')
    @mock.patch('cinder.objects.GroupList.get_all_by_project')
    def test_get_all(self, is_admin, mock_get_all_by_project,
                     mock_get_all):
        self.group_api.LOG = mock.Mock()
        fake_groups = ['fake_group1', 'fake_group2']
        fake_groups_by_project = ['fake_group1']
        mock_get_all.return_value = fake_groups
        mock_get_all_by_project.return_value = fake_groups_by_project

        if is_admin:
            grps = self.group_api.get_all(self.ctxt,
                                          filters={'all_tenants': True})
            self.assertEqual(fake_groups, grps)
        else:
            grps = self.group_api.get_all(self.user_ctxt)
            self.assertEqual(fake_groups_by_project, grps)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.delete_group')
    @mock.patch('cinder.db.volume_get_all_by_generic_group')
    @mock.patch('cinder.db.volumes_update')
    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.db.group_type_get')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_create_delete(self, mock_volume_types_get,
                           mock_group_type_get, mock_group,
                           mock_update_quota, mock_cast_create_group,
                           mock_volumes_update, mock_volume_get_all,
                           mock_rpc_delete_group):
        mock_volume_types_get.return_value = [{'id': fake.VOLUME_TYPE_ID}]
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova', host=None,
                                 name=name, description=description,
                                 status=fields.GroupStatus.CREATING)
        mock_group.return_value = grp

        ret_group = self.group_api.create(self.ctxt, name, description,
                                          fake.GROUP_TYPE_ID,
                                          [fake.VOLUME_TYPE_ID],
                                          availability_zone='nova')
        self.assertEqual(grp.obj_to_primitive(), ret_group.obj_to_primitive())

        ret_group.host = "test_host@fakedrv#fakepool"
        ret_group.status = fields.GroupStatus.AVAILABLE
        ret_group.assert_not_frozen = mock.Mock(return_value=True)
        ret_group.group_snapshots = []
        self.group_api.delete(self.ctxt, ret_group, delete_volumes=True)
        mock_volume_get_all.assert_called_once_with(mock.ANY, ret_group.id)
        mock_volumes_update.assert_called_once_with(self.ctxt, [])
        mock_rpc_delete_group.assert_called_once_with(self.ctxt, ret_group)

    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.db.group_type_get_by_name')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_create_with_group_name(self, mock_volume_types_get,
                                    mock_group_type_get, mock_group,
                                    mock_update_quota, mock_cast_create_group):
        mock_volume_types_get.return_value = [{'id': fake.VOLUME_TYPE_ID}]
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova', host=None,
                                 name=name, description=description,
                                 status=fields.GroupStatus.CREATING)
        mock_group.return_value = grp

        ret_group = self.group_api.create(self.ctxt, name, description,
                                          "fake-grouptype-name",
                                          [fake.VOLUME_TYPE_ID],
                                          availability_zone='nova')
        self.assertEqual(grp.obj_to_primitive(), ret_group.obj_to_primitive())

        mock_group_type_get.assert_called_once_with(self.ctxt,
                                                    "fake-grouptype-name")

    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.db.group_type_get')
    @mock.patch('cinder.db.group_type_get_by_name')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_create_with_uuid_format_group_type_name(
            self, mock_volume_types_get, mock_group_type_get_by_name,
            mock_group_type_get, mock_update_quota, mock_cast_create_group):
        uuid_format_type_name = fake.UUID1
        mock_volume_types_get.return_value = [{'id': fake.VOLUME_TYPE_ID}]
        mock_group_type_get.side_effect = exception.GroupTypeNotFound(
            group_type_id=uuid_format_type_name)
        mock_group_type_get_by_name.return_value = {'id': fake.GROUP_TYPE_ID}

        ret_group = self.group_api.create(self.ctxt, "test_group", '',
                                          uuid_format_type_name,
                                          [fake.VOLUME_TYPE_ID],
                                          availability_zone='nova')
        self.assertEqual(ret_group["group_type_id"],
                         fake.GROUP_TYPE_ID)

    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.db.group_type_get_by_name')
    @mock.patch('cinder.db.sqlalchemy.api._volume_type_get')
    @mock.patch('cinder.db.sqlalchemy.api._volume_type_get_by_name')
    def test_create_with_uuid_format_volume_type_name(
            self, mock_vol_t_get_by_name, mock_vol_types_get_by_id,
            mock_group_type_get, mock_update_quota, mock_cast_create_group):
        uuid_format_name = fake.UUID1
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        volume_type = {'id': fake.VOLUME_TYPE_ID, 'name': uuid_format_name}
        mock_vol_types_get_by_id.side_effect = exception.VolumeTypeNotFound(
            volume_type_id=uuid_format_name)
        mock_vol_t_get_by_name.return_value = volume_type
        group = self.group_api.create(self.ctxt, "test_group",
                                      "this is a test group",
                                      "fake-grouptype-name",
                                      [uuid_format_name],
                                      availability_zone='nova')
        self.assertEqual(group["volume_type_ids"],
                         [volume_type['id']])

    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.db.group_type_get_by_name')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_create_with_multi_types(self, mock_volume_types_get,
                                     mock_group_type_get,
                                     mock_update_quota,
                                     mock_cast_create_group):
        volume_types = [{'id': fake.VOLUME_TYPE_ID},
                        {'id': fake.VOLUME_TYPE2_ID}]
        mock_volume_types_get.return_value = volume_types
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        volume_type_names = ['fake-volume-type1', 'fake-volume-type2']
        name = "test_group"
        description = "this is a test group"

        group = self.group_api.create(self.ctxt, name, description,
                                      "fake-grouptype-name",
                                      volume_type_names,
                                      availability_zone='nova')
        self.assertEqual(group["volume_type_ids"],
                         [t['id'] for t in volume_types])
        self.assertEqual(group["group_type_id"], fake.GROUP_TYPE_ID)

        mock_group_type_get.assert_called_once_with(self.ctxt,
                                                    "fake-grouptype-name")
        mock_volume_types_get.assert_called_once_with(mock.ANY,
                                                      volume_type_names)

    @mock.patch('oslo_utils.timeutils.utcnow')
    @mock.patch('cinder.objects.Group')
    def test_reset_status(self, mock_group, mock_time_util):
        mock_time_util.return_value = "time_now"
        self.group_api.reset_status(self.ctxt, mock_group,
                                    fields.GroupStatus.AVAILABLE)

        update_field = {'updated_at': "time_now",
                        'status': fields.GroupStatus.AVAILABLE}
        mock_group.update.assert_called_once_with(update_field)
        mock_group.save.assert_called_once_with()

    @mock.patch.object(GROUP_QUOTAS, "reserve")
    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.db.group_type_get_by_name')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_create_group_failed_update_quota(self,
                                              mock_volume_types_get,
                                              mock_group_type_get, mock_group,
                                              mock_group_quota_reserve):
        mock_volume_types_get.return_value = [{'id': fake.VOLUME_TYPE_ID}]
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        fake_overs = ['groups']
        fake_quotas = {'groups': 1}
        fake_usages = {'groups': {'reserved': 0, 'in_use': 1}}
        mock_group_quota_reserve.side_effect = exception.OverQuota(
            overs=fake_overs,
            quotas=fake_quotas,
            usages=fake_usages)
        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova', host=None,
                                 name=name, description=description,
                                 status=fields.GroupStatus.CREATING)
        mock_group.return_value = grp

        self.assertRaises(exception.GroupLimitExceeded,
                          self.group_api.create,
                          self.ctxt, name, description,
                          "fake-grouptype-name",
                          [fake.VOLUME_TYPE_ID],
                          availability_zone='nova')

    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.db.volume_get')
    def test__validate_add_volumes(self, mock_volume_get, mock_group):
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova', host=None,
                                 name="name", description="description",
                                 status=fields.GroupStatus.CREATING)
        mock_group.return_value = grp
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctxt)
        mock_volume_get.return_value = fake_volume_obj
        self.assertRaises(exception.InvalidVolume,
                          self.group_api._validate_add_volumes, self.ctxt,
                          [], ['123456789'], grp)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.update_group')
    @mock.patch('cinder.db.volume_get_all_by_generic_group')
    @mock.patch('cinder.group.api.API._cast_create_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.db.group_type_get')
    @mock.patch('cinder.db.volume_types_get_by_name_or_id')
    def test_update(self, mock_volume_types_get,
                    mock_group_type_get, mock_group,
                    mock_update_quota, mock_cast_create_group,
                    mock_volume_get_all, mock_rpc_update_group):
        vol_type_dict = {'id': fake.VOLUME_TYPE_ID,
                         'name': 'fake_volume_type'}
        vol_type = objects.VolumeType(self.ctxt, **vol_type_dict)

        mock_volume_types_get.return_value = [{'id': fake.VOLUME_TYPE_ID}]
        mock_group_type_get.return_value = {'id': fake.GROUP_TYPE_ID}
        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova', host=None,
                                 name=name, description=description,
                                 status=fields.GroupStatus.CREATING)
        mock_group.return_value = grp

        ret_group = self.group_api.create(self.ctxt, name, description,
                                          fake.GROUP_TYPE_ID,
                                          [fake.VOLUME_TYPE_ID],
                                          availability_zone='nova')
        self.assertEqual(grp.obj_to_primitive(), ret_group.obj_to_primitive())

        ret_group.volume_types = [vol_type]
        ret_group.host = "test_host@fakedrv#fakepool"
        ret_group.status = fields.GroupStatus.AVAILABLE
        ret_group.id = fake.GROUP_ID

        vol1 = utils.create_volume(
            self.ctxt, host=ret_group.host,
            availability_zone=ret_group.availability_zone,
            volume_type_id=fake.VOLUME_TYPE_ID)

        vol2 = utils.create_volume(
            self.ctxt, host=ret_group.host,
            availability_zone=ret_group.availability_zone,
            volume_type_id=fake.VOLUME_TYPE_ID,
            group_id=fake.GROUP_ID)
        vol2_dict = {
            'id': vol2.id,
            'group_id': fake.GROUP_ID,
            'volume_type_id': fake.VOLUME_TYPE_ID,
            'availability_zone': ret_group.availability_zone,
            'host': ret_group.host,
            'status': 'available',
        }
        mock_volume_get_all.return_value = [vol2_dict]

        new_name = "new_group_name"
        new_desc = "this is a new group"
        self.group_api.update(self.ctxt, ret_group, new_name, new_desc,
                              vol1.id, vol2.id)
        mock_volume_get_all.assert_called_once_with(mock.ANY, ret_group.id)
        mock_rpc_update_group.assert_called_once_with(self.ctxt, ret_group,
                                                      add_volumes=vol1.id,
                                                      remove_volumes=vol2.id)

    @mock.patch('cinder.objects.GroupSnapshot.get_by_id')
    @mock.patch('cinder.context.RequestContext.authorize')
    def test_get_group_snapshot(self, mock_authorize, mock_group_snap):
        fake_group_snap = 'fake_group_snap'
        mock_group_snap.return_value = fake_group_snap
        grp_snap = self.group_api.get_group_snapshot(
            self.ctxt, fake.GROUP_SNAPSHOT_ID)
        self.assertEqual(fake_group_snap, grp_snap)
        mock_authorize.assert_called_once_with(
            g_snap_policies.GET_POLICY,
            target_obj=fake_group_snap)

    @ddt.data(True, False)
    @mock.patch('cinder.objects.GroupSnapshotList.get_all')
    @mock.patch('cinder.objects.GroupSnapshotList.get_all_by_project')
    def test_get_all_group_snapshots(self, is_admin,
                                     mock_get_all_by_project,
                                     mock_get_all):
        fake_group_snaps = ['fake_group_snap1', 'fake_group_snap2']
        fake_group_snaps_by_project = ['fake_group_snap1']
        mock_get_all.return_value = fake_group_snaps
        mock_get_all_by_project.return_value = fake_group_snaps_by_project

        if is_admin:
            grp_snaps = self.group_api.get_all_group_snapshots(
                self.ctxt, filters={'all_tenants': True})
            self.assertEqual(fake_group_snaps, grp_snaps)
        else:
            grp_snaps = self.group_api.get_all_group_snapshots(
                self.user_ctxt)
            self.assertEqual(fake_group_snaps_by_project, grp_snaps)

    @mock.patch('cinder.objects.GroupSnapshot')
    def test_update_group_snapshot(self, mock_group_snap):
        grp_snap_update = {"name": "new_name",
                           "description": "This is a new description"}
        self.group_api.update_group_snapshot(self.ctxt, mock_group_snap,
                                             grp_snap_update)
        mock_group_snap.update.assert_called_once_with(grp_snap_update)
        mock_group_snap.save.assert_called_once_with()

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.delete_group_snapshot')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_group_snapshot')
    @mock.patch('cinder.volume.api.API.create_snapshots_in_db')
    @mock.patch('cinder.objects.Group')
    @mock.patch('cinder.objects.GroupSnapshot')
    @mock.patch('cinder.objects.SnapshotList.get_all_for_group_snapshot')
    def test_create_delete_group_snapshot(self,
                                          mock_snap_get_all,
                                          mock_group_snap, mock_group,
                                          mock_create_in_db,
                                          mock_create_api, mock_delete_api):
        name = "fake_name"
        description = "fake description"
        mock_group.id = fake.GROUP_ID
        mock_group.group_type_id = fake.GROUP_TYPE_ID
        mock_group.assert_not_frozen = mock.Mock(return_value=True)
        mock_group.volumes = []
        ret_group_snap = self.group_api.create_group_snapshot(
            self.ctxt, mock_group, name, description)
        mock_snap_get_all.return_value = []

        options = {'group_id': fake.GROUP_ID,
                   'user_id': self.ctxt.user_id,
                   'project_id': self.ctxt.project_id,
                   'status': "creating",
                   'name': name,
                   'description': description,
                   'group_type_id': fake.GROUP_TYPE_ID}
        mock_group_snap.assert_called_once_with(self.ctxt, **options)
        ret_group_snap.create.assert_called_once_with()
        mock_create_in_db.assert_called_once_with(self.ctxt, [],
                                                  ret_group_snap.name,
                                                  ret_group_snap.description,
                                                  None,
                                                  ret_group_snap.id)
        mock_create_api.assert_called_once_with(self.ctxt, ret_group_snap)

        ret_group_snap.assert_not_frozen = mock.Mock(return_value=True)
        self.group_api.delete_group_snapshot(self.ctxt, ret_group_snap)
        mock_delete_api.assert_called_once_with(mock.ANY, ret_group_snap)

    @mock.patch('cinder.volume.api.API.delete')
    @mock.patch('cinder.objects.VolumeType.get_by_name_or_id')
    @mock.patch('cinder.db.group_volume_type_mapping_create')
    @mock.patch('cinder.volume.api.API.create')
    @mock.patch('cinder.objects.GroupSnapshot.get_by_id')
    @mock.patch('cinder.objects.SnapshotList.get_all_for_group_snapshot')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_group_from_src')
    @mock.patch('cinder.objects.VolumeList.get_all_by_generic_group')
    def test_create_group_from_snap_volume_failed(
            self, mock_volume_get_all,
            mock_rpc_create_group_from_src,
            mock_snap_get_all, mock_group_snap_get,
            mock_volume_api_create,
            mock_mapping_create,
            mock_get_volume_type, mock_volume_delete):
        mock_volume_api_create.side_effect = [exception.CinderException]
        vol_type = fake_volume.fake_volume_type_obj(
            self.ctxt,
            id=fake.VOLUME_TYPE_ID,
            name='fake_volume_type')
        mock_get_volume_type.return_value = vol_type

        grp_snap = utils.create_group_snapshot(
            self.ctxt, fake.GROUP_ID,
            group_type_id=fake.GROUP_TYPE_ID,
            status=fields.GroupStatus.CREATING)
        mock_group_snap_get.return_value = grp_snap
        vol1 = utils.create_volume(
            self.ctxt,
            availability_zone='nova',
            volume_type_id=vol_type['id'],
            group_id=fake.GROUP_ID)

        snap = utils.create_snapshot(self.ctxt, vol1.id,
                                     volume_type_id=vol_type['id'],
                                     status=fields.GroupStatus.CREATING)
        mock_snap_get_all.return_value = [snap]

        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[vol_type['id']],
                                 availability_zone='nova',
                                 name=name, description=description,
                                 group_snapshot_id=grp_snap.id,
                                 status=fields.GroupStatus.CREATING)

        vol2 = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=vol_type['id'],
            group_id=grp.id,
            snapshot_id=snap.id)
        mock_volume_get_all.return_value = [vol2]

        self.assertRaises(
            exception.CinderException,
            self.group_api._create_group_from_group_snapshot,
            self.ctxt, grp, grp_snap.id)

        mock_volume_api_create.assert_called_once_with(
            self.ctxt, 1, None, None,
            availability_zone=grp.availability_zone,
            group_snapshot=grp_snap,
            group=grp,
            snapshot=snap,
            volume_type=vol_type)

        mock_rpc_create_group_from_src.assert_not_called()

        mock_volume_delete.assert_called_once_with(self.ctxt, vol2)

        vol2.destroy()
        grp.destroy()
        snap.destroy()
        vol1.destroy()
        grp_snap.destroy()

    @mock.patch('cinder.objects.VolumeType.get_by_name_or_id')
    @mock.patch('cinder.db.group_volume_type_mapping_create')
    @mock.patch('cinder.volume.api.API.create')
    @mock.patch('cinder.objects.GroupSnapshot.get_by_id')
    @mock.patch('cinder.objects.SnapshotList.get_all_for_group_snapshot')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_group_from_src')
    @mock.patch('cinder.objects.VolumeList.get_all_by_generic_group')
    def test_create_group_from_snap(self, mock_volume_get_all,
                                    mock_rpc_create_group_from_src,
                                    mock_snap_get_all, mock_group_snap_get,
                                    mock_volume_api_create,
                                    mock_mapping_create,
                                    mock_get_volume_type):
        vol_type = fake_volume.fake_volume_type_obj(
            self.ctxt,
            id=fake.VOLUME_TYPE_ID,
            name='fake_volume_type')
        mock_get_volume_type.return_value = vol_type

        grp_snap = utils.create_group_snapshot(
            self.ctxt, fake.GROUP_ID,
            group_type_id=fake.GROUP_TYPE_ID,
            status=fields.GroupStatus.CREATING)
        mock_group_snap_get.return_value = grp_snap
        vol1 = utils.create_volume(
            self.ctxt,
            availability_zone='nova',
            volume_type_id=vol_type['id'],
            group_id=fake.GROUP_ID)

        snap = utils.create_snapshot(self.ctxt, vol1.id,
                                     volume_type_id=vol_type['id'],
                                     status=fields.GroupStatus.CREATING)
        mock_snap_get_all.return_value = [snap]

        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[vol_type['id']],
                                 availability_zone='nova',
                                 name=name, description=description,
                                 group_snapshot_id=grp_snap.id,
                                 status=fields.GroupStatus.CREATING)

        vol2 = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=vol_type['id'],
            group_id=grp.id,
            snapshot_id=snap.id)
        mock_volume_get_all.return_value = [vol2]

        self.group_api._create_group_from_group_snapshot(self.ctxt, grp,
                                                         grp_snap.id)

        mock_volume_api_create.assert_called_once_with(
            self.ctxt, 1, None, None,
            availability_zone=grp.availability_zone,
            group_snapshot=grp_snap,
            group=grp,
            snapshot=snap,
            volume_type=vol_type)

        mock_rpc_create_group_from_src.assert_called_once_with(
            self.ctxt, grp, grp_snap)

        vol2.destroy()
        grp.destroy()
        snap.destroy()
        vol1.destroy()
        grp_snap.destroy()

    @mock.patch('cinder.objects.VolumeType.get_by_name_or_id')
    @mock.patch('cinder.db.group_volume_type_mapping_create')
    @mock.patch('cinder.volume.api.API.create')
    @mock.patch('cinder.objects.Group.get_by_id')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_group_from_src')
    @mock.patch('cinder.objects.VolumeList.get_all_by_generic_group')
    def test_create_group_from_group(self, mock_volume_get_all,
                                     mock_rpc_create_group_from_src,
                                     mock_group_get,
                                     mock_volume_api_create,
                                     mock_mapping_create,
                                     mock_get_volume_type):
        vol_type = fake_volume.fake_volume_type_obj(
            self.ctxt,
            id=fake.VOLUME_TYPE_ID,
            name='fake_volume_type')
        mock_get_volume_type.return_value = vol_type

        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[vol_type['id']],
                                 availability_zone='nova',
                                 status=fields.GroupStatus.CREATING)
        mock_group_get.return_value = grp

        vol = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=fake.VOLUME_TYPE_ID,
            group_id=grp.id)
        mock_volume_get_all.return_value = [vol]

        grp2 = utils.create_group(self.ctxt,
                                  group_type_id=fake.GROUP_TYPE_ID,
                                  volume_type_ids=[vol_type['id']],
                                  availability_zone='nova',
                                  source_group_id=grp.id,
                                  status=fields.GroupStatus.CREATING)

        vol2 = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=vol_type['id'],
            group_id=grp2.id,
            source_volid=vol.id)

        self.group_api._create_group_from_source_group(self.ctxt, grp2,
                                                       grp.id)

        mock_volume_api_create.assert_called_once_with(
            self.ctxt, 1, None, None,
            availability_zone=grp.availability_zone,
            source_group=grp,
            group=grp2,
            source_volume=vol,
            volume_type=vol_type)

        mock_rpc_create_group_from_src.assert_called_once_with(
            self.ctxt, grp2, None, grp)

        vol2.destroy()
        grp2.destroy()
        vol.destroy()
        grp.destroy()

    @mock.patch('cinder.volume.api.API.delete')
    @mock.patch('cinder.objects.VolumeType.get_by_name_or_id')
    @mock.patch('cinder.db.group_volume_type_mapping_create')
    @mock.patch('cinder.volume.api.API.create')
    @mock.patch('cinder.objects.Group.get_by_id')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_group_from_src')
    @mock.patch('cinder.objects.VolumeList.get_all_by_generic_group')
    def test_create_group_from_group_create_volume_failed(
            self, mock_volume_get_all, mock_rpc_create_group_from_src,
            mock_group_get, mock_volume_api_create, mock_mapping_create,
            mock_get_volume_type, mock_volume_delete):
        vol_type = fake_volume.fake_volume_type_obj(
            self.ctxt,
            id=fake.VOLUME_TYPE_ID,
            name='fake_volume_type')
        mock_get_volume_type.return_value = vol_type

        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[vol_type['id']],
                                 availability_zone='nova',
                                 status=fields.GroupStatus.CREATING)
        mock_group_get.return_value = grp

        vol1 = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=fake.VOLUME_TYPE_ID,
            group_id=grp.id)
        vol2 = utils.create_volume(
            self.ctxt,
            availability_zone=grp.availability_zone,
            volume_type_id=fake.VOLUME_TYPE_ID,
            group_id=grp.id)
        mock_volume_get_all.side_effect = [[vol1, vol2], [vol1]]

        grp2 = utils.create_group(self.ctxt,
                                  group_type_id=fake.GROUP_TYPE_ID,
                                  volume_type_ids=[vol_type['id']],
                                  availability_zone='nova',
                                  source_group_id=grp.id,
                                  status=fields.GroupStatus.CREATING)

        mock_volume_api_create.side_effect = [None, exception.CinderException]

        self.assertRaises(
            exception.CinderException,
            self.group_api._create_group_from_source_group,
            self.ctxt, grp2, grp.id)

        mock_rpc_create_group_from_src.assert_not_called()
        mock_volume_delete.assert_called_once_with(self.ctxt, vol1)

        grp2.destroy()
        vol2.destroy()
        vol1.destroy()
        grp.destroy()

    @mock.patch('cinder.group.api.API._create_group_from_group_snapshot')
    @mock.patch('cinder.group.api.API._create_group_from_source_group')
    @mock.patch('cinder.group.api.API.update_quota')
    @mock.patch('cinder.objects.GroupSnapshot.get_by_id')
    @mock.patch('cinder.objects.SnapshotList.get_all_for_group_snapshot')
    @mock.patch('cinder.scheduler.rpcapi.SchedulerAPI.validate_host_capacity')
    def test_create_from_src(self, mock_validate_host, mock_snap_get_all,
                             mock_group_snap_get, mock_update_quota,
                             mock_create_from_group,
                             mock_create_from_snap):
        name = "test_group"
        description = "this is a test group"
        grp = utils.create_group(self.ctxt, group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[fake.VOLUME_TYPE_ID],
                                 availability_zone='nova',
                                 name=name, description=description,
                                 status=fields.GroupStatus.AVAILABLE,)

        vol1 = utils.create_volume(
            self.ctxt,
            availability_zone='nova',
            volume_type_id=fake.VOLUME_TYPE_ID,
            group_id=grp.id)

        snap = utils.create_snapshot(self.ctxt, vol1.id,
                                     volume_type_id=fake.VOLUME_TYPE_ID,
                                     status=fields.SnapshotStatus.AVAILABLE)
        mock_snap_get_all.return_value = [snap]
        mock_validate_host.return_host = True

        grp_snap = utils.create_group_snapshot(
            self.ctxt, grp.id,
            group_type_id=fake.GROUP_TYPE_ID,
            status=fields.GroupStatus.AVAILABLE)
        mock_group_snap_get.return_value = grp_snap

        grp2 = utils.create_group(self.ctxt,
                                  group_type_id=fake.GROUP_TYPE_ID,
                                  volume_type_ids=[fake.VOLUME_TYPE_ID],
                                  availability_zone='nova',
                                  name=name, description=description,
                                  status=fields.GroupStatus.CREATING,
                                  group_snapshot_id=grp_snap.id)

        with mock.patch('cinder.objects.Group') as mock_group:
            mock_group.return_value = grp2
            with mock.patch('cinder.objects.group.Group.create'):
                ret_group = self.group_api.create_from_src(
                    self.ctxt, name, description,
                    group_snapshot_id=grp_snap.id,
                    source_group_id=None)
                self.assertEqual(grp2.obj_to_primitive(),
                                 ret_group.obj_to_primitive())
                mock_create_from_snap.assert_called_once_with(
                    self.ctxt, grp2, grp_snap.id)

        snap.destroy()
        grp_snap.destroy()
        vol1.destroy()
        grp.destroy()
        grp2.destroy()

    @mock.patch('oslo_utils.timeutils.utcnow')
    @mock.patch('cinder.objects.GroupSnapshot')
    def test_reset_group_snapshot_status(self,
                                         mock_group_snapshot,
                                         mock_time_util):
        mock_time_util.return_value = "time_now"
        self.group_api.reset_group_snapshot_status(
            self.ctxt, mock_group_snapshot, fields.GroupSnapshotStatus.ERROR)

        update_field = {'updated_at': "time_now",
                        'status': fields.GroupSnapshotStatus.ERROR}
        mock_group_snapshot.update.assert_called_once_with(update_field)
        mock_group_snapshot.save.assert_called_once_with()

    @mock.patch('cinder.scheduler.rpcapi.SchedulerAPI.validate_host_capacity')
    def test_create_group_from_src_frozen(self, mock_validate_host):
        service = utils.create_service(self.ctxt, {'frozen': True})
        group = utils.create_group(self.ctxt, host=service.host,
                                   group_type_id='gt')
        mock_validate_host.return_value = True
        group_api = cinder.group.api.API()
        self.assertRaises(exception.InvalidInput,
                          group_api.create_from_src,
                          self.ctxt, 'group', 'desc',
                          group_snapshot_id=None, source_group_id=group.id)

    def test_delete_group_frozen(self):
        service = utils.create_service(self.ctxt, {'frozen': True})
        group = utils.create_group(self.ctxt, host=service.host,
                                   group_type_id='gt')
        group_api = cinder.group.api.API()
        self.assertRaises(exception.InvalidInput,
                          group_api.delete, self.ctxt, group)

    def test_create_group_snapshot_frozen(self):
        service = utils.create_service(self.ctxt, {'frozen': True})
        group = utils.create_group(self.ctxt, host=service.host,
                                   group_type_id='gt')
        group_api = cinder.group.api.API()
        self.assertRaises(exception.InvalidInput,
                          group_api.create_group_snapshot,
                          self.ctxt, group, 'group_snapshot', 'desc')

    def test_delete_group_snapshot_frozen(self):
        service = utils.create_service(self.ctxt, {'frozen': True})
        group = utils.create_group(self.ctxt, host=service.host,
                                   group_type_id='gt')
        gsnap = utils.create_group_snapshot(self.ctxt, group.id)
        group_api = cinder.group.api.API()
        self.assertRaises(exception.InvalidInput,
                          group_api.delete_group_snapshot,
                          self.ctxt, gsnap)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs',
                return_value={'qos_specs': {}})
    @mock.patch('cinder.scheduler.rpcapi.SchedulerAPI.create_group')
    def test_cast_create_group(self,
                               mock_create_group,
                               mock_get_volume_type_qos_specs):
        vol_type = utils.create_volume_type(self.ctxt, name='test_vol_type')
        encryption_key_id = mock.sentinel.encryption_key_id
        description = mock.sentinel.description
        name = mock.sentinel.name
        req_spec = {'volume_type': vol_type,
                    'encryption_key_id': encryption_key_id,
                    'description': description,
                    'name': name}

        grp_name = "test_group"
        grp_description = "this is a test group"
        grp_spec = {'name': grp_name,
                    'description': grp_description}

        grp = utils.create_group(self.ctxt,
                                 group_type_id=fake.GROUP_TYPE_ID,
                                 volume_type_ids=[vol_type.id],
                                 availability_zone='nova')

        grp_filter_properties = mock.sentinel.group_filter_properties
        filter_properties_list = mock.sentinel.filter_properties_list
        self.group_api._cast_create_group(self.ctxt,
                                          grp,
                                          grp_spec,
                                          [req_spec],
                                          grp_filter_properties,
                                          filter_properties_list)

        mock_get_volume_type_qos_specs.assert_called_once_with(vol_type.id)

        exp_vol_properties = {
            'size': 0,
            'user_id': self.ctxt.user_id,
            'project_id': self.ctxt.project_id,
            'status': 'creating',
            'attach_status': 'detached',
            'encryption_key_id': encryption_key_id,
            'display_description': description,
            'display_name': name,
            'volume_type_id': vol_type.id,
            'group_type_id': grp.group_type_id,
            'availability_zone': grp.availability_zone
        }
        exp_req_spec = {
            'volume_type': vol_type,
            'encryption_key_id': encryption_key_id,
            'description': description,
            'name': name,
            'volume_properties': exp_vol_properties,
            'qos_specs': None
        }
        exp_grp_properties = {
            'size': 0,
            'user_id': self.ctxt.user_id,
            'project_id': self.ctxt.project_id,
            'status': 'creating',
            'display_description': grp_description,
            'display_name': grp_name,
            'group_type_id': grp.group_type_id,
        }
        exp_grp_spec = {
            'name': grp_name,
            'description': grp_description,
            'volume_properties': exp_grp_properties,
            'qos_specs': None
        }
        mock_create_group.assert_called_once_with(
            self.ctxt,
            grp,
            group_spec=exp_grp_spec,
            request_spec_list=[exp_req_spec],
            group_filter_properties=grp_filter_properties,
            filter_properties_list=filter_properties_list)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.tests.functional.api_sample_tests import test_volumes


class VolumeSnapshotsSampleJsonTest(test_volumes.VolumesSampleBase):
    sample_dir = "snapshots"

    def setUp(self):
        super(VolumeSnapshotsSampleJsonTest, self).setUp()
        res = self._create_volume()
        res = jsonutils.loads(res.content)['volume']
        self._poll_volume_while(res['id'], ['creating'])
        self.subs = {
            "volume_id": res['id']
        }
        self.response = self._create_snapshot(self.subs)

    def _create_snapshot(self, subs=None):
        response = self._do_post('snapshots',
                                 'snapshot-create-request',
                                 subs)
        return response

    def test_snapshot_list_detail(self):

        response = self._do_get('snapshots/detail')
        self._verify_response('snapshots-list-detailed-response',
                              {}, response, 200)

    def test_snapshot_create(self):

        self._verify_response('snapshot-create-response',
                              {}, self.response, 202)

    def test_snapshot_list(self):

        response = self._do_get('snapshots')
        self._verify_response('snapshots-list-response',
                              {}, response, 200)

    def test_snapshot_metadata_show(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s/metadata' % res['id'])
        self._verify_response('snapshot-metadata-show-response',
                              {}, response, 200)

    def test_snapshot_metadata_create(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_post('snapshots/%s/metadata' % res['id'],
                                 'snapshot-metadata-create-request')
        self._verify_response('snapshot-metadata-create-response',
                              {}, response, 200)

    def test_snapshot_metadata_update(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_put('snapshots/%s/metadata' % res['id'],
                                'snapshot-metadata-update-request')
        self._verify_response('snapshot-metadata-update-response',
                              {}, response, 200)

    def test_snapshot_show(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s' % res['id'])
        self._verify_response('snapshot-show-response',
                              {}, response, 200)

    def test_snapshot_update(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_put('snapshots/%s' % res['id'],
                                'snapshot-update-request')
        self._verify_response('snapshot-update-response',
                              {}, response, 200)

    def test_snapshot_metadata_show_specific_key(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_get('snapshots/%s/metadata/key' % res['id'])
        self._verify_response('snapshot-metadata-show-key-response',
                              {}, response, 200)

    def test_snapshot_metadata_update_specific_key(self):

        res = jsonutils.loads(self.response.content)['snapshot']
        response = self._do_put('snapshots/%s/metadata/key' % res['id'],
                                'snapshot-metadata-update-key-request')
        self._verify_response('snapshot-metadata-update-key-response',
                              {}, response, 200)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from cinder.policies import base


CREATE_POLICY = "volume:create_transfer"
ACCEPT_POLICY = "volume:accept_transfer"
DELETE_POLICY = "volume:delete_transfer"
GET_POLICY = "volume:get_transfer"
GET_ALL_POLICY = "volume:get_all_transfers"


volume_transfer_policies = [
    policy.DocumentedRuleDefault(
        name=GET_ALL_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="List volume transfer.",
        operations=[
            {
                'method': 'GET',
                'path': '/os-volume-transfer'
            },
            {
                'method': 'GET',
                'path': '/os-volume-transfer/detail'
            },
            {
                'method': 'GET',
                'path': '/volume_transfers'
            },
            {
                'method': 'GET',
                'path': '/volume-transfers/detail'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=CREATE_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Create a volume transfer.",
        operations=[
            {
                'method': 'POST',
                'path': '/os-volume-transfer'
            },
            {
                'method': 'POST',
                'path': '/volume_transfers'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=GET_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Show one specified volume transfer.",
        operations=[
            {
                'method': 'GET',
                'path': '/os-volume-transfer/{transfer_id}'
            },
            {
                'method': 'GET',
                'path': '/volume-transfers/{transfer_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=ACCEPT_POLICY,
        check_str="",
        description="Accept a volume transfer.",
        operations=[
            {
                'method': 'POST',
                'path': '/os-volume-transfer/{transfer_id}/accept'
            },
            {
                'method': 'POST',
                'path': '/volume-transfers/{transfer_id}/accept'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=DELETE_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Delete volume transfer.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/os-volume-transfer/{transfer_id}'
            },
            {
                'method': 'DELETE',
                'path': '/volume-transfers/{transfer_id}'
            }
        ]),
]


def list_rules():
    return volume_transfer_policies



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2016 Vedams Inc.
# Copyright (C) 2016 Google Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for Google Backup code.

"""

import bz2
import filecmp
import hashlib
import os
import shutil
import tempfile
import threading
import zlib

from eventlet import tpool
import mock
from oslo_utils import units

from cinder.backup.drivers import gcs as google_dr
from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder import test
from cinder.tests.unit.backup import fake_google_client
from cinder.tests.unit.backup import fake_google_client2
from cinder.tests.unit import fake_constants as fake


class FakeMD5(object):
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def digest(cls):
        return 'gcscindermd5'

    @classmethod
    def hexdigest(cls):
        return 'gcscindermd5'


class FakeObjectName(object):
    @classmethod
    def _fake_generate_object_name_prefix(cls, backup):
        az = 'az_fake'
        backup_name = '%s_backup_%s' % (az, backup.id)
        volume = 'volume_%s' % (backup.volume_id)
        prefix = volume + '_' + backup_name
        return prefix


def gcs_client(func):
    @mock.patch.object(google_dr.client, 'GoogleCredentials',
                       fake_google_client.FakeGoogleCredentials)
    @mock.patch.object(google_dr.discovery, 'build',
                       fake_google_client.FakeGoogleDiscovery.Build)
    @mock.patch.object(google_dr, 'GoogleMediaIoBaseDownload',
                       fake_google_client.FakeGoogleMediaIoBaseDownload)
    @mock.patch.object(hashlib, 'md5', FakeMD5)
    def func_wrapper(self, *args, **kwargs):
        if google_dr.service_account:
            with mock.patch.object(google_dr.service_account.Credentials,
                                   'from_service_account_file',
                                   fake_google_client.FakeGoogleCredentials):
                return func(self, *args, **kwargs)
        return func(self, *args, **kwargs)

    return func_wrapper


def gcs_client2(func):
    @mock.patch.object(google_dr.client, 'GoogleCredentials',
                       fake_google_client2.FakeGoogleCredentials)
    @mock.patch.object(google_dr.discovery, 'build',
                       fake_google_client2.FakeGoogleDiscovery.Build)
    @mock.patch.object(google_dr, 'GoogleMediaIoBaseDownload',
                       fake_google_client2.FakeGoogleMediaIoBaseDownload)
    @mock.patch.object(google_dr.GoogleBackupDriver,
                       '_generate_object_name_prefix',
                       FakeObjectName._fake_generate_object_name_prefix)
    @mock.patch.object(hashlib, 'md5', FakeMD5)
    def func_wrapper(self, *args, **kwargs):
        if google_dr.service_account:
            with mock.patch.object(google_dr.service_account.Credentials,
                                   'from_service_account_file',
                                   fake_google_client.FakeGoogleCredentials):
                return func(self, *args, **kwargs)
        return func(self, *args, **kwargs)

    return func_wrapper


def fake_backup_metadata(self, backup, object_meta):
    raise exception.BackupDriverException(reason=_('fake'))


def fake_delete(self, backup):
    raise exception.BackupOperationError()


def _fake_delete_object(self, bucket_name, object_name):
    raise AssertionError('delete_object method should not be called.')


class GoogleBackupDriverTestCase(test.TestCase):
    """Test Case for Google"""

    _DEFAULT_VOLUME_ID = 'c7eb81f4-bec6-4730-a60f-8888885874df'

    def _create_volume_db_entry(self, volume_id=_DEFAULT_VOLUME_ID):
        vol = {'id': volume_id,
               'size': 1,
               'status': 'available',
               'volume_type_id': self.vt['id']}
        return db.volume_create(self.ctxt, vol)['id']

    def _create_backup_db_entry(self,
                                volume_id=_DEFAULT_VOLUME_ID,
                                container=google_dr.CONF.backup_gcs_bucket,
                                parent_id=None,
                                status=None,
                                service_metadata=None):

        try:
            db.volume_get(self.ctxt, volume_id)
        except exception.NotFound:
            self._create_volume_db_entry(volume_id=volume_id)

        kwargs = {'size': 1,
                  'container': container,
                  'volume_id': volume_id,
                  'parent_id': parent_id,
                  'user_id': fake.USER_ID,
                  'project_id': fake.PROJECT_ID,
                  'status': status,
                  'service_metadata': service_metadata,
                  }
        backup = objects.Backup(context=self.ctxt, **kwargs)
        backup.create()
        return backup

    def _write_effective_compression_file(self, data_size):
        """Ensure file contents can be effectively compressed."""
        self.volume_file.seek(0)
        self.volume_file.write(bytes([65] * data_size))
        self.volume_file.seek(0)

    def setUp(self):
        super(GoogleBackupDriverTestCase, self).setUp()
        self.flags(backup_gcs_bucket='gcscinderbucket')
        self.flags(backup_gcs_credential_file='test-file')
        self.flags(backup_gcs_project_id='test-gcs')
        self.ctxt = context.get_admin_context()
        self.volume_file = tempfile.NamedTemporaryFile()
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(self.volume_file.close)
        # Remove tempdir.
        self.addCleanup(shutil.rmtree, self.temp_dir)
        self.size_volume_file = 0
        for _i in range(0, 64):
            self.volume_file.write(os.urandom(units.Ki))
            self.size_volume_file += 1024
        # Note(yikun): It mocks out the backup notifier to avoid to leak
        # notifications into other test.
        notify_patcher = mock.patch(
            'cinder.volume.volume_utils.notify_about_backup_usage')
        notify_patcher.start()
        self.addCleanup(notify_patcher.stop)

    @gcs_client
    def test_backup(self):
        volume_id = 'b09b1ad4-5f0e-4d3f-8b9e-0000004f5ec2'
        container_name = 'test-bucket'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        result = service.backup(backup, self.volume_file)
        self.assertIsNone(result)

    @gcs_client
    def test_backup_uncompressed(self):
        volume_id = '2b9f10a3-42b4-4fdf-b316-000000ceb039'
        backup = self._create_backup_db_entry(volume_id=volume_id)
        self.flags(backup_compression_algorithm='none')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)

    @gcs_client
    def test_backup_bz2(self):
        volume_id = 'dc0fee35-b44e-4f13-80d6-000000e1b50c'
        backup = self._create_backup_db_entry(volume_id=volume_id)
        self.flags(backup_compression_algorithm='bz2')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self._write_effective_compression_file(self.size_volume_file)
        service.backup(backup, self.volume_file)

    @gcs_client
    def test_backup_zlib(self):
        volume_id = '5cea0535-b6fb-4531-9a38-000000bea094'
        backup = self._create_backup_db_entry(volume_id=volume_id)
        self.flags(backup_compression_algorithm='zlib')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self._write_effective_compression_file(self.size_volume_file)
        service.backup(backup, self.volume_file)

    @gcs_client
    def test_backup_default_container(self):
        volume_id = '9552017f-c8b9-4e4e-a876-00000053349c'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=None)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertEqual('gcscinderbucket', backup.container)

    @gcs_client
    @mock.patch('httplib2.proxy_info_from_url')
    def test_backup_proxy_configured(self, mock_proxy_info):
        # Configuration overwrites enviromental variable
        proxy_cfg = "http://myproxy.example.com"
        os.environ['http_proxy'] = proxy_cfg + '_fake'
        google_dr.CONF.set_override("backup_gcs_proxy_url", proxy_cfg)
        google_dr.GoogleBackupDriver(self.ctxt)
        self.assertEqual(proxy_cfg, os.environ.get('http_proxy'))

    @gcs_client
    @mock.patch('cinder.backup.drivers.gcs.GoogleBackupDriver.'
                '_send_progress_end')
    @mock.patch('cinder.backup.drivers.gcs.GoogleBackupDriver.'
                '_send_progress_notification')
    def test_backup_default_container_notify(self, _send_progress,
                                             _send_progress_end):
        volume_id = '87dd0eed-2598-4ebd-8ebb-000000ac578a'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=None)
        # If the backup_object_number_per_notification is set to 1,
        # the _send_progress method will be called for sure.
        google_dr.CONF.set_override("backup_object_number_per_notification", 1)
        google_dr.CONF.set_override("backup_gcs_enable_progress_timer", False)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertTrue(_send_progress.called)
        self.assertTrue(_send_progress_end.called)

        # If the backup_object_number_per_notification is increased to
        # another value, the _send_progress method will not be called.
        _send_progress.reset_mock()
        _send_progress_end.reset_mock()
        google_dr.CONF.set_override("backup_object_number_per_notification",
                                    10)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertFalse(_send_progress.called)
        self.assertTrue(_send_progress_end.called)

        # If the timer is enabled, the _send_progress will be called,
        # since the timer can trigger the progress notification.
        _send_progress.reset_mock()
        _send_progress_end.reset_mock()
        google_dr.CONF.set_override("backup_object_number_per_notification",
                                    10)
        google_dr.CONF.set_override("backup_gcs_enable_progress_timer", True)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertTrue(_send_progress.called)
        self.assertTrue(_send_progress_end.called)

    @gcs_client
    def test_backup_custom_container(self):
        volume_id = '1da9859e-77e5-4731-bd58-000000ca119e'
        container_name = 'fake99'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertEqual(container_name, backup.container)

    @gcs_client2
    def test_backup_shafile(self):
        volume_id = '6465dad4-22af-48f7-8a1a-000000218907'

        container_name = self.temp_dir.replace(tempfile.gettempdir() + '/',
                                               '', 1)
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service.backup(backup, self.volume_file)
        self.assertEqual(container_name, backup.container)

        # Verify sha contents
        content1 = service._read_sha256file(backup)
        self.assertEqual(64 * units.Ki / content1['chunk_size'],
                         len(content1['sha256s']))

    @gcs_client2
    def test_backup_cmp_shafiles(self):
        volume_id = '1a99ac67-c534-4fe3-b472-0000001785e2'

        container_name = self.temp_dir.replace(tempfile.gettempdir() + '/',
                                               '', 1)
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service1 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service1.backup(backup, self.volume_file)
        self.assertEqual(container_name, backup.container)

        # Create incremental backup with no change to contents
        deltabackup = self._create_backup_db_entry(volume_id=volume_id,
                                                   container=container_name,
                                                   parent_id=backup.id)
        service2 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service2.backup(deltabackup, self.volume_file)
        self.assertEqual(container_name, deltabackup.container)

        # Compare shas from both files
        content1 = service1._read_sha256file(backup)
        content2 = service2._read_sha256file(deltabackup)

        self.assertEqual(len(content1['sha256s']), len(content2['sha256s']))
        self.assertEqual(set(content1['sha256s']), set(content2['sha256s']))

    @gcs_client2
    def test_backup_delta_two_objects_change(self):
        volume_id = '30dab288-265a-4583-9abe-000000d42c67'

        self.flags(backup_gcs_object_size=8 * units.Ki)
        self.flags(backup_gcs_block_size=units.Ki)

        container_name = self.temp_dir.replace(tempfile.gettempdir() + '/',
                                               '', 1)
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service1 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service1.backup(backup, self.volume_file)
        self.assertEqual(container_name, backup.container)

        # Create incremental backup with no change to contents
        self.volume_file.seek(2 * 8 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))
        self.volume_file.seek(4 * 8 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))

        deltabackup = self._create_backup_db_entry(volume_id=volume_id,
                                                   container=container_name,
                                                   parent_id=backup.id)
        service2 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service2.backup(deltabackup, self.volume_file)
        self.assertEqual(container_name, deltabackup.container)

        content1 = service1._read_sha256file(backup)
        content2 = service2._read_sha256file(deltabackup)

        # Verify that two shas are changed at index 16 and 32
        self.assertNotEqual(content1['sha256s'][16], content2['sha256s'][16])
        self.assertNotEqual(content1['sha256s'][32], content2['sha256s'][32])

    @gcs_client2
    def test_backup_delta_two_blocks_in_object_change(self):
        volume_id = 'b943e84f-aa67-4331-9ab2-000000cf19ba'

        self.flags(backup_gcs_object_size=8 * units.Ki)
        self.flags(backup_gcs_block_size=units.Ki)

        container_name = self.temp_dir.replace(tempfile.gettempdir() + '/',
                                               '', 1)
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)

        service1 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service1.backup(backup, self.volume_file)
        self.assertEqual(container_name, backup.container)

        # Create incremental backup with no change to contents
        self.volume_file.seek(16 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))
        self.volume_file.seek(20 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))

        deltabackup = self._create_backup_db_entry(volume_id=volume_id,
                                                   container=container_name,
                                                   parent_id=backup.id)
        service2 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service2.backup(deltabackup, self.volume_file)
        self.assertEqual(container_name, deltabackup.container)

        # Verify that two shas are changed at index 16 and 20
        content1 = service1._read_sha256file(backup)
        content2 = service2._read_sha256file(deltabackup)
        self.assertNotEqual(content1['sha256s'][16], content2['sha256s'][16])
        self.assertNotEqual(content1['sha256s'][20], content2['sha256s'][20])

    @gcs_client
    def test_create_backup_fail(self):
        volume_id = 'b09b1ad4-5f0e-4d3f-8b9e-0000004f5ec3'
        container_name = 'gcs_api_failure'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        self.assertRaises(google_dr.GCSApiFailure,
                          service.backup,
                          backup, self.volume_file)

    @gcs_client
    def test_create_backup_fail2(self):
        volume_id = 'b09b1ad4-5f0e-4d3f-8b9e-0000004f5ec4'
        container_name = 'gcs_oauth2_failure'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        self.assertRaises(google_dr.GCSOAuth2Failure,
                          service.backup,
                          backup, self.volume_file)

    @gcs_client
    @mock.patch.object(google_dr.GoogleBackupDriver, '_backup_metadata',
                       fake_backup_metadata)
    def test_backup_backup_metadata_fail(self):
        """Test of when an exception occurs in backup().

        In backup(), after an exception occurs in
        self._backup_metadata(), we want to check the process of an
        exception handler.
        """
        volume_id = '020d9142-339c-4876-a445-000000f1520c'

        backup = self._create_backup_db_entry(volume_id=volume_id)
        self.flags(backup_compression_algorithm='none')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        # We expect that an exception be notified directly.
        self.assertRaises(exception.BackupDriverException,
                          service.backup,
                          backup, self.volume_file)

    @gcs_client
    @mock.patch.object(google_dr.GoogleBackupDriver, '_backup_metadata',
                       fake_backup_metadata)
    @mock.patch.object(google_dr.GoogleBackupDriver, 'delete_backup',
                       fake_delete)
    def test_backup_backup_metadata_fail2(self):
        """Test of when an exception occurs in an exception handler.

        In backup(), after an exception occurs in
        self._backup_metadata(), we want to check the process when the
        second exception occurs in self.delete_backup().
        """
        volume_id = '2164421d-f181-4db7-b9bd-000000eeb628'

        backup = self._create_backup_db_entry(volume_id=volume_id)
        self.flags(backup_compression_algorithm='none')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        # We expect that the second exception is notified.
        self.assertRaises(exception.BackupOperationError,
                          service.backup,
                          backup, self.volume_file)

    @gcs_client
    def test_restore(self):
        volume_id = 'c2a81f09-f480-4325-8424-00000071685b'
        backup = self._create_backup_db_entry(
            volume_id=volume_id,
            status=objects.fields.BackupStatus.RESTORING)
        service = google_dr.GoogleBackupDriver(self.ctxt)

        with tempfile.NamedTemporaryFile() as volume_file:
            service.restore(backup, volume_id, volume_file)

    @gcs_client
    def test_restore_fail(self):
        volume_id = 'c2a81f09-f480-4325-8424-00000071685b'
        container_name = 'gcs_connection_failure'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service = google_dr.GoogleBackupDriver(self.ctxt)

        with tempfile.NamedTemporaryFile() as volume_file:
            self.assertRaises(google_dr.GCSConnectionFailure,
                              service.restore,
                              backup, volume_id, volume_file)

    @gcs_client2
    def test_restore_delta(self):
        volume_id = '04d83506-bcf7-4ff5-9c65-00000051bd2e'
        self.flags(backup_gcs_object_size=8 * units.Ki)
        self.flags(backup_gcs_block_size=units.Ki)
        container_name = self.temp_dir.replace(tempfile.gettempdir() + '/',
                                               '', 1)
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              container=container_name)
        service1 = google_dr.GoogleBackupDriver(self.ctxt)
        self.volume_file.seek(0)
        service1.backup(backup, self.volume_file)

        # Create incremental backup with no change to contents
        self.volume_file.seek(16 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))
        self.volume_file.seek(20 * units.Ki)
        self.volume_file.write(os.urandom(units.Ki))

        deltabackup = self._create_backup_db_entry(
            volume_id=volume_id,
            status=objects.fields.BackupStatus.RESTORING,
            container=container_name,
            parent_id=backup.id)
        self.volume_file.seek(0)
        service2 = google_dr.GoogleBackupDriver(self.ctxt)
        service2.backup(deltabackup, self.volume_file, True)

        with tempfile.NamedTemporaryFile() as restored_file:
            service2.restore(deltabackup, volume_id,
                             restored_file)
            self.assertTrue(filecmp.cmp(self.volume_file.name,
                            restored_file.name))

    @gcs_client
    def test_delete(self):
        volume_id = '9ab256c8-3175-4ad8-baa1-0000007f9d31'
        object_prefix = 'test_prefix'
        backup = self._create_backup_db_entry(volume_id=volume_id,
                                              service_metadata=object_prefix)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        service.delete_backup(backup)

    @gcs_client
    @mock.patch.object(google_dr.GoogleBackupDriver, 'delete_object',
                       _fake_delete_object)
    def test_delete_without_object_prefix(self):
        volume_id = 'ee30d649-72a6-49a5-b78d-000000edb6b1'
        backup = self._create_backup_db_entry(volume_id=volume_id)
        service = google_dr.GoogleBackupDriver(self.ctxt)
        service.delete_backup(backup)

    @gcs_client
    def test_get_compressor(self):
        service = google_dr.GoogleBackupDriver(self.ctxt)
        compressor = service._get_compressor('None')
        self.assertIsNone(compressor)
        compressor = service._get_compressor('zlib')
        self.assertEqual(zlib, compressor)
        self.assertIsInstance(compressor, tpool.Proxy)
        compressor = service._get_compressor('bz2')
        self.assertEqual(bz2, compressor)
        self.assertIsInstance(compressor, tpool.Proxy)
        self.assertRaises(ValueError, service._get_compressor, 'fake')

    @gcs_client
    def test_prepare_output_data_effective_compression(self):
        """Test compression works on a native thread."""
        # Use dictionary to share data between threads
        thread_dict = {}
        original_compress = zlib.compress

        def my_compress(data):
            thread_dict['compress'] = threading.current_thread()
            return original_compress(data)

        self.mock_object(zlib, 'compress', side_effect=my_compress)

        service = google_dr.GoogleBackupDriver(self.ctxt)
        # Set up buffer of 128 zeroed bytes
        fake_data = b'\0' * 128

        result = service._prepare_output_data(fake_data)

        self.assertEqual('zlib', result[0])
        self.assertGreater(len(fake_data), len(result[1]))
        self.assertNotEqual(threading.current_thread(),
                            thread_dict['compress'])

    @gcs_client
    def test_prepare_output_data_no_compression(self):
        self.flags(backup_compression_algorithm='none')
        service = google_dr.GoogleBackupDriver(self.ctxt)
        # Set up buffer of 128 zeroed bytes
        fake_data = b'\0' * 128

        result = service._prepare_output_data(fake_data)

        self.assertEqual('none', result[0])
        self.assertEqual(fake_data, result[1])

    @gcs_client
    def test_prepare_output_data_ineffective_compression(self):
        service = google_dr.GoogleBackupDriver(self.ctxt)
        # Set up buffer of 128 zeroed bytes
        fake_data = b'\0' * 128
        # Pre-compress so that compression in the driver will be ineffective.
        already_compressed_data = service.compressor.compress(fake_data)

        result = service._prepare_output_data(already_compressed_data)

        self.assertEqual('none', result[0])
        self.assertEqual(already_compressed_data, result[1])

    @mock.patch('googleapiclient.__version__', '1.5.5')
    @mock.patch.object(google_dr.client.GoogleCredentials, 'from_stream')
    @mock.patch.object(google_dr.discovery, 'build')
    @mock.patch.object(google_dr, 'service_account')
    def test_non_google_auth_version(self, account, build, from_stream):
        # Prior to v1.6.0 Google api client doesn't support google-auth library
        google_dr.CONF.set_override('backup_gcs_credential_file',
                                    'credentials_file')

        google_dr.GoogleBackupDriver(self.ctxt)

        from_stream.assert_called_once_with('credentials_file')
        account.Credentials.from_service_account_file.assert_not_called()
        build.assert_called_once_with('storage', 'v1', cache_discovery=False,
                                      credentials=from_stream.return_value)

    @mock.patch('googleapiclient.__version__', '1.6.6')
    @mock.patch.object(google_dr.client.GoogleCredentials, 'from_stream')
    @mock.patch.object(google_dr.discovery, 'build')
    @mock.patch.object(google_dr, 'service_account', None)
    def test_no_httplib2_auth(self, build, from_stream):
        # Google api client requires google-auth-httplib2 if not present we
        # use legacy credentials
        google_dr.CONF.set_override('backup_gcs_credential_file',
                                    'credentials_file')

        google_dr.GoogleBackupDriver(self.ctxt)

        from_stream.assert_called_once_with('credentials_file')
        build.assert_called_once_with('storage', 'v1', cache_discovery=False,
                                      credentials=from_stream.return_value)

    @mock.patch('googleapiclient.__version__', '1.6.6')
    @mock.patch.object(google_dr, 'gexceptions', mock.Mock())
    @mock.patch.object(google_dr.client.GoogleCredentials, 'from_stream')
    @mock.patch.object(google_dr.discovery, 'build')
    @mock.patch.object(google_dr, 'service_account')
    def test_google_auth_used(self, account, build, from_stream):
        # Google api client requires google-auth-httplib2 if not present we
        # use legacy credentials
        google_dr.CONF.set_override('backup_gcs_credential_file',
                                    'credentials_file')

        google_dr.GoogleBackupDriver(self.ctxt)

        from_stream.assert_not_called()
        create_creds = account.Credentials.from_service_account_file
        create_creds.assert_called_once_with('credentials_file')
        build.assert_called_once_with('storage', 'v1', cache_discovery=False,
                                      credentials=create_creds.return_value)



#    Copyright 2014 Objectif Libre
#    Copyright 2015 Dot Hill Systems Corp.
#    Copyright 2016-2019 Seagate Technology or one of its affiliates
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
import cinder.volume.driver
import cinder.volume.drivers.san.san as san
import cinder.volume.drivers.stx.common as common
import cinder.volume.drivers.stx.exception as stx_exception


DEFAULT_ISCSI_PORT = "3260"
LOG = logging.getLogger(__name__)


class STXISCSIDriver(cinder.volume.driver.ISCSIDriver):
    """OpenStack iSCSI Cinder driver for Seagate storage arrays.

    .. code:: text

      Version history:
          0.1    - Base structure for DotHill iSCSI drivers based on HPMSA FC
                   drivers:
                       "https://github.com/openstack/cinder/tree/stable/juno/
                        cinder/volume/drivers/san/hp"
          1.0    - Version developed for DotHill arrays with the following
                   modifications:
                       - added iSCSI support
                       - added CHAP support in iSCSI
                       - added support for v3 API(virtual pool feature)
                       - added support for retype volume
                       - added support for manage/unmanage volume
                       - added https support
          1.6    - Add management path redundancy and reduce load placed
                   on management controller.
          1.7    - Modified so it can't be invoked except as a superclass
          2.0    - Reworked to create a new Seagate (STX) array driver.
    """

    VERSION = "2.0"

    CI_WIKI_NAME = 'Seagate_CI'

    def __init__(self, *args, **kwargs):
        super(STXISCSIDriver, self).__init__(*args, **kwargs)
        self.common = None
        self.configuration.append_config_values(san.san_opts)
        if type(self) != STXISCSIDriver:
            return
        self.configuration.append_config_values(common.common_opts)
        self.configuration.append_config_values(common.iscsi_opts)
        self.iscsi_ips = self.configuration.seagate_iscsi_ips

    def _init_common(self):
        return common.STXCommon(self.configuration)

    def _check_flags(self):
        required_flags = ['san_ip', 'san_login', 'san_password']
        self.common.check_flags(self.configuration, required_flags)

    def do_setup(self, context):
        self.common = self._init_common()
        self._check_flags()
        self.common.do_setup(context)
        self.initialize_iscsi_ports()

    def initialize_iscsi_ports(self):
        iscsi_ips = []
        if self.iscsi_ips:
            for ip_addr in self.iscsi_ips:
                ip = ip_addr.split(':')
                if len(ip) == 1:
                    iscsi_ips.append([ip_addr, DEFAULT_ISCSI_PORT])
                elif len(ip) == 2:
                    iscsi_ips.append([ip[0], ip[1]])
                else:
                    msg = _("Invalid IP address format: '%s'") % ip_addr
                    LOG.error(msg)
                    raise exception.InvalidInput(reason=(msg))
            self.iscsi_ips = iscsi_ips
        else:
            msg = _('At least one valid iSCSI IP address must be set.')
            LOG.error(msg)
            raise exception.InvalidInput(reason=(msg))

    def check_for_setup_error(self):
        self._check_flags()

    def create_volume(self, volume):
        self.common.create_volume(volume)

    def create_volume_from_snapshot(self, volume, src_vref):
        self.common.create_volume_from_snapshot(volume, src_vref)

    def create_cloned_volume(self, volume, src_vref):
        self.common.create_cloned_volume(volume, src_vref)

    def delete_volume(self, volume):
        self.common.delete_volume(volume)

    def initialize_connection(self, volume, connector):
        self.common.client_login()
        try:
            data = {}
            data['target_lun'] = self.common.map_volume(volume,
                                                        connector,
                                                        'initiator')
            iqns = self.common.get_active_iscsi_target_iqns()
            data['target_discovered'] = True
            data['target_iqn'] = iqns[0]
            iscsi_portals = self.common.get_active_iscsi_target_portals()

            for ip_port in self.iscsi_ips:
                if (ip_port[0] in iscsi_portals):
                    data['target_portal'] = ":".join(ip_port)
                    break

            if 'target_portal' not in data:
                raise stx_exception.NotTargetPortal()

            if self.configuration.use_chap_auth:
                chap_secret = self.common.get_chap_record(
                    connector['initiator']
                )
                if not chap_secret:
                    chap_secret = self.create_chap_record(
                        connector['initiator']
                    )
                data['auth_password'] = chap_secret
                data['auth_username'] = connector['initiator']
                data['auth_method'] = 'CHAP'

            info = {'driver_volume_type': 'iscsi',
                    'data': data}
            return info
        finally:
            self.common.client_logout()

    def terminate_connection(self, volume, connector, **kwargs):
        if type(connector) == dict and 'initiator' in connector:
            # multiattach volumes cannot be unmapped here, but will
            # be implicity unmapped when the volume is deleted.
            if not volume.get('multiattach'):
                self.common.unmap_volume(volume, connector, 'initiator')

    def get_volume_stats(self, refresh=False):
        stats = self.common.get_volume_stats(refresh)
        stats['storage_protocol'] = 'iSCSI'
        stats['driver_version'] = self.VERSION
        backend_name = self.configuration.safe_get('volume_backend_name')
        stats['volume_backend_name'] = (backend_name or
                                        self.__class__.__name__)
        return stats

    def create_export(self, context, volume, connector=None):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    def create_snapshot(self, snapshot):
        self.common.create_snapshot(snapshot)

    def delete_snapshot(self, snapshot):
        self.common.delete_snapshot(snapshot)

    def extend_volume(self, volume, new_size):
        self.common.extend_volume(volume, new_size)

    def create_chap_record(self, initiator_name):
        chap_secret = self.configuration.chap_password
        # Chap secret length should be 12 to 16 characters
        if 12 <= len(chap_secret) <= 16:
            self.common.create_chap_record(initiator_name, chap_secret)
        else:
            msg = _('CHAP secret should be 12-16 bytes.')
            LOG.error(msg)
            raise exception.InvalidInput(reason=(msg))
        return chap_secret

    def retype(self, context, volume, new_type, diff, host):
        return self.common.retype(volume, new_type, diff, host)

    def manage_existing(self, volume, existing_ref):
        self.common.manage_existing(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        return self.common.manage_existing_get_size(volume, existing_ref)

    def unmanage(self, volume):
        pass
